# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# ogbujipt.text_helper

'''
Routines to help with text processing
'''
import re
import warnings


def text_splitter(text: str,
        chunk_size: int,
        chunk_overlap: int=0,
        separator: str='\n\n',
        len_func=len
    ) -> list[str]:
    '''
    Split string into a set of chunks

    Note that this function is "fuzzy"; it will not necessarily split the text into perfectly chunk_size length chunks, 
    and will preserve items between the given separators. this results in slightly larger chunks than requested.

    Much like langchain's CharTextSplitter.py

    >>> from ogbujipt.text_helper import text_splitter
    >>> from PyPDF2 import PdfReader
    >>> pdf_reader = PdfReader('monopoly-board-game-manual.pdf')
    >>> text = ''.join((page.extract_text() for page in pdf_reader.pages))
    >>> chunks = text_splitter(text, chunk_size=500, separator='\n')

    Args:
        text (str): (Multiline) String to be split into chunks

        chunk_size (int): Number of characters to include per chunk

        chunk_overlap (int, optional): Number of characters to overlap at the edges of chunks

        seperator (str, optional): String that already splits "text" into sections

        distance_function (callable, optional): Function to measure length, len() by default

    Returns:
        chunks (List[str]): List of chunks of the text provided
    '''
    assert separator, 'Separator must be non-empty'
    
    if ((not isinstance(text, str))
        or (not isinstance(separator, str))):
        raise ValueError(f'text and separator must be strings.\n'
                         f'Got {text.__class__} for text and {separator.__class__} for separator')
    
    if ((not isinstance(chunk_size, int))
        or (not isinstance(chunk_overlap, int))
        or (chunk_size <= 0)
        or (chunk_overlap < 0)
        or (chunk_size < chunk_overlap)):
        raise ValueError(f'chunk_size must be a positive integer, '
                         f'chunk_overlap must be a non-negative integer, and'
                         f'chunk_size must be greater than chunk_overlap.\n'
                         f'Got {chunk_size} chunk_size and {chunk_overlap} chunk_overlap.')
    
    # Split up the text by the separator
    # FIXME: Need a step for escaping regex
    sep_pat = re.compile(separator)
    fine_split = re.split(sep_pat, text)
    separator_len = len_func(separator)

    if len(fine_split) <= 1:
        warnings.warn(f'No splits detected. Problem with separator ({repr(separator)})?')

    # Combine the small pieces into medium size chunks to send to LLM
    # Initialize accumulators; chunks will be the target list of the chunks so far
    # curr_chunk will be a list of parts comprising the main, current chunk
    # back_overlap will be appended, once ready, to the end of the previous chunk (if any)
    # fwd_overlap will be prepended, once ready, to the start of the next chunk
    chunks = []
    curr_chunk, curr_chunk_len = [], 0
    back_overlap, back_overlap_len = None, 0  # None signals not yet gathering
    fwd_overlap, fwd_overlap_len = None, 0

    for s in fine_split:
        if not s: continue  # noqa E701
        split_len = len_func(s) + separator_len
        # Check for full back_overlap (if relevant, i.e. back_overlap isn't None)
        if back_overlap is not None and (back_overlap_len + split_len > chunk_overlap):  # noqa: F821
            chunks[-1].extend(back_overlap)
            back_overlap, back_overlap_len = None, 0

        # Will adding this split take us into overlap room?
        if curr_chunk_len + split_len > (chunk_size - chunk_overlap):
            fwd_overlap, fwd_overlap_len = [], 0  # Start gathering

        # Will adding this split take us over chunk size?
        if curr_chunk_len + split_len > chunk_size:
            # If so, complete current chunk & start a new one

            # fwd_overlap should be non-None at this point, so check empty
            if not fwd_overlap and curr_chunk:
                # If empty, look back to make sure there is some overlap
                fwd_overlap.append(curr_chunk[-1])

            chunks.append(curr_chunk)
            # fwd_overlap intentionally not counted in running chunk length
            curr_chunk, curr_chunk_len = fwd_overlap, 0
            back_overlap, back_overlap_len = [], 0  # Start gathering
            fwd_overlap, fwd_overlap_len = None, 0  # Stop gathering

        if fwd_overlap is not None:
            fwd_overlap.append(s)
            fwd_overlap_len += split_len

        if back_overlap is not None:
            back_overlap.append(s)
            back_overlap_len += split_len

        curr_chunk.append(s)
        curr_chunk_len += split_len

    # Done with the splits; use the final back_overlap, if any
    if back_overlap:
        chunks[-1].extend(back_overlap)

    # Concatenate all the split parts of all the chunks
    chunks = [separator.join(c) for c in chunks]

    # Handle degenerate case where no splits found & chunk size too large
    # Just becomes one big chunk
    if not chunks:
        chunks = [text]

    # chunks.append(separator.join(curr_chunk))
    return chunks



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
#
# SPDX-License-Identifier: Apache-2.0
# ogbujipt

# ruff: noqa: F401,F403

from .__about__ import __version__


def oapi_first_choice_text(response):
    '''
    Given an OpenAI-compatible API simple completion response, return the first choice text
    '''
    import warnings
    warnings.warn(DeprecationWarning('Use the openai_api class\'s first_choice_text() method instead'))
    return response['choices'][0]['text']


def oapi_chat_first_choice_message(response):
    '''
    Given an OpenAI-compatible API chat completion response, return the first choice message content
    '''
    import warnings
    warnings.warn(DeprecationWarning('Use the openai_chat_api class\'s first_choice_message() method instead'))
    return response['choices'][0]['message']['content']



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# test/embedding/test_pgvector_data.py
'''
See test/embedding/test_pgvector.py for important notes on running these tests
'''

import pytest
from unittest.mock import MagicMock, DEFAULT  # noqa: F401

import numpy as np


KG_STATEMENTS = [  # Demo data
    ("👤 Alikiba `releases_single` 💿 'Yalaiti'", {'url': 'https://notjustok.com/lyrics/yalaiti-lyrics-by-alikiba-ft-sabah-salum/'}),
    ("👤 Sabah Salum `featured_in` 💿 'Yalaiti'", {'url': 'https://notjustok.com/lyrics/yalaiti-lyrics-by-alikiba-ft-sabah-salum/'}),
    ('👤 Kukbeatz `collaborates_with` 👤 Ruger', {'url': 'https://notjustok.com/lyrics/all-of-us-lyrics-by-kukbeatz-ft-ruger/'}),
    ('💿 All of Us `is_a_song_by` 👤 Kukbeatz and Ruger', {'url': 'https://notjustok.com/lyrics/all-of-us-lyrics-by-kukbeatz-ft-ruger/'}),
    ('👤 Blaqbonez `collaborates_with` 👤 Fireboy DML', {'url': 'https://notjustok.com/news/snippet-of-fireboy-dmls-collaboration-with-blaqbonez/'})
]


# XXX: Move to a fixture?
# Definitely don't want to even import SentenceTransformer class due to massive side-effects
class SentenceTransformer(object):
    def __init__(self, model_name_or_path):
        self.encode = MagicMock()


@pytest.mark.asyncio
async def test_insert_data_vector(DB):
    dummy_model = SentenceTransformer('mock_transformer')
    dummy_model.encode.return_value = np.array([1, 2, 3])

    item1_text = KG_STATEMENTS[0][0]
    item1_meta = KG_STATEMENTS[0][1]

    # Insert data
    for index, (text, meta) in enumerate(KG_STATEMENTS):
        await DB.insert(                                  # Insert the row into the table
            content=text,                                  # text to be embedded
            tags=[f'{k}={v}' for (k, v) in meta.items()],  # Tag metadata
        )

    assert await DB.count_items() == len(KG_STATEMENTS), Exception('Incorrect number of documents after insertion')

    # search table with perfect match
    result = await DB.search(text=item1_text, limit=3)
    # assert result is not None, Exception('No results returned from perfect search')

    # Even though the embedding is mocked, the stored text should be faithful
    row = next(result)
    assert row.content == item1_text
    assert row.tags == [f'{k}={v}' for (k, v) in item1_meta.items()]

    await DB.drop_table()


@pytest.mark.asyncio
async def test_insertmany_data_vector(DB):
    dummy_model = SentenceTransformer('mock_transformer')
    dummy_model.encode.return_value = np.array([1, 2, 3])

    item1_text = KG_STATEMENTS[0][0]
    item1_meta = KG_STATEMENTS[0][1]

    # Insert data using insert_many()
    dataset = ((text, [f'{k}={v}' for (k, v) in tags.items()]) for (text, tags) in KG_STATEMENTS)
    
    await DB.insert_many(dataset)

    assert await DB.count_items() == len(KG_STATEMENTS), Exception('Incorrect number of documents after insertion')

    # search table with perfect match
    result = await DB.search(text=item1_text, limit=3)
    # assert result is not None, Exception('No results returned from perfect search')

    # Even though the embedding is mocked, the stored text should be faithful
    row = next(result)
    assert row.content == item1_text
    assert row.tags == [f'{k}={v}' for (k, v) in item1_meta.items()]

    await DB.drop_table()


if __name__ == '__main__':
    raise SystemExit("Attention! Run with pytest")



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
#
# SPDX-License-Identifier: Apache-2.0
# ogbujipt.prompting

# ruff: noqa: F401,F403

from ogbujipt.prompting.basic import format, pdelim, ordering
from ogbujipt.prompting.model_style import *  # noqa F403 Yes that module has an __all__



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# ogbujipt.embedding.pgvector

'''
Vector databases embeddings using PGVector
'''

import json
# import asyncio
# from typing import ClassVar

from ogbujipt.config import attr_dict

# Handle key imports
try:
    import asyncpg
    from pgvector.asyncpg import register_vector
    PREREQS_AVAILABLE = True
    POOL_TYPE = asyncpg.pool.Pool
except ImportError:
    PREREQS_AVAILABLE = False
    asyncpg = None
    register_vector = object()  # Set up a dummy to satisfy the type hints
    POOL_TYPE = object

# ------ SQL queries ---------------------------------------------------------------------------------------------------
# PG only supports proper query arguments (e.g. $1, $2, etc.) for values, not for table or column names
# Table names are checked to be legit sequel table names, and embed_dimension is assured to be an integer

CREATE_VECTOR_EXTENSION = 'CREATE EXTENSION IF NOT EXISTS vector;'

CHECK_TABLE_EXISTS = '''-- Check if a table exists
SELECT EXISTS (
    SELECT FROM pg_tables
    WHERE tablename = $1
);
'''
# ------ SQL queries ---------------------------------------------------------------------------------------------------

# Default overall PG max is 100.
# See: https://commandprompt.com/education/how-to-alter-max_connections-parameter-in-postgresql/
DEFAULT_MIN_MAX_CONNECTION_POOL_SIZE = (10, 20)


class PGVectorHelper:
    '''
    Helper class for PGVector operations

    Construct using PGVectorHelper.from_conn_params() method

    Connection and pool parameters:

    * table_name: PostgresQL table name. Checked to restrict to alphanumeric characters & underscore
    * host: Hostname or IP address of the PostgreSQL server. Defaults to UNIX socket if not provided.
    * port: Port number at which the PostgreSQL server is listening. Defaults to 5432 if not provided.
    * user: User name used to authenticate.
    * password: Password used to authenticate.
    * database: Database name to connect to.
    * min_max_size: Tuple of minimum and maximum number of connections to maintain in the pool.
        Defaults to (10, 20)
    '''
    def __init__(self, embedding_model, table_name: str, pool: asyncpg.pool.Pool):
        '''
        If you don't already have a connection pool, construct using the PGvectorHelper.from_pool_params() method

        Args:
            embedding (SentenceTransformer): SentenceTransformer object of your choice
            https://huggingface.co/sentence-transformers

            table_name: PostgresQL table. Checked to restrict to alphanumeric characters & underscore

            pool: asyncpg connection pool instance
        '''
        if not PREREQS_AVAILABLE:
            raise RuntimeError('pgvector not installed, you can run `pip install pgvector asyncpg`')

        if not table_name.replace('_', '').isalnum():
            msg = 'table_name must be alphanumeric, with underscore also allowed'
            raise ValueError(msg)

        self.table_name = table_name
        self.embedding_model = embedding_model
        self.pool = pool

        # Check if the provided embedding model is a SentenceTransformer
        if (embedding_model.__class__.__name__ == 'SentenceTransformer') and (not None):
            self._embedding_model = embedding_model
            self._embed_dimension = len(self._embedding_model.encode(''))
        elif embedding_model is None:
            self._embedding_model = None
            self._embed_dimension = 0
        else:
            raise ValueError('embedding_model must be a SentenceTransformer object or None')

        self.table_name = table_name
        self.pool = pool

    @classmethod
    async def from_conn_params(cls, embedding_model, table_name, host, port, db_name, user, password) -> 'PGVectorHelper': # noqa: E501
        '''
        Create PGVectorHelper instance from connection/pool parameters

        Will create a connection pool for you, with JSON type handling initialized,
        and set that as a pool attribute on the created object as a user convenience.

        For details on accepted parameters, See the class docstring
            (e.g. run `help(PGVectorHelper)`)
        '''
        pool = await asyncpg.create_pool(init=PGVectorHelper.init_pool, host=host, port=port, user=user,
                                        password=password, database=db_name)

        new_obj = cls(embedding_model, table_name, pool)
        return new_obj

    @staticmethod
    async def init_pool(conn):
        '''
        Initialize vector extension for a connection from a pool
        '''
        await conn.execute('CREATE EXTENSION IF NOT EXISTS vector;')
        await register_vector(conn)
        await conn.set_type_codec(  # Register a codec for JSON
            'JSON', encoder=json.dumps, decoder=json.loads, schema='pg_catalog')

    # Hmm. Just called count in the qdrant version
    async def count_items(self) -> int:
        '''
        Count the number of documents in the table

        Returns:
            int: number of documents in the table
        '''
        async with self.pool.acquire() as conn:
            # Count the number of documents in the table
            count = await conn.fetchval(f'SELECT COUNT(*) FROM {self.table_name}')
        return count

    async def table_exists(self) -> bool:
        '''
        Check if the table exists

        Returns:
            bool: True if the table exists, False otherwise
        '''
        # Check if the table exists
        async with self.pool.acquire() as conn:
            exists = await conn.fetchval(
                CHECK_TABLE_EXISTS,
                self.table_name
            )
        return exists

    async def drop_table(self) -> None:
        '''
        Delete the table

        Exercise caution!
        '''
        # Delete the table
        async with self.pool.acquire() as conn:
            await conn.execute(f'DROP TABLE IF EXISTS {self.table_name};')


def process_search_response(qresponse):
    '''
    Convert a query response to an attributable dict

    Args:
        query_response (asyncpg.Record): asyncpg.Record object to be converted to a dict

    Returns:
        list[dict]: List with a dict representation for each result row

    >>> await mydb.search(text='Hello')
    >>> results = process_search_response()
    >>> row = next(results)  # Assume there's at least one result
    >>> c = r.content
    >>> t = r.tags

    If a row does not have a title or page_numbers field, these will be set to None

    Other reasons for this conversion: asyncpg.Record objects are not JSON serializable,
    and don't support attribute-style access
    '''
    for row in qresponse:
        # Actually, this is otiose; just let missing attributes fail
        # if 'title' not in row:
        #     row['title'] = None
        # if 'page_numbers' not in row:
        #     row['page_numbers'] = None
        # print(row, row.items())
        yield attr_dict(row)


# Down here to avoid circular imports
from ogbujipt.embedding.pgvector_data_doc import DataDB, DocDB  # noqa: E402 F401
from ogbujipt.embedding.pgvector_message import MessageDB  # noqa: E402 F401



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
#
# SPDX-License-Identifier: Apache-2.0

# XXX: I think this test has an __init__.py as part of hatch's
# developer/live environments scheme? If this is the case,
# it's not long for this world. We just use pytest directly.



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# ogbujipt.config

# Given the nature of a config module, we don't use __all__
# & avoid top-level imports where possible, and at the end of the module
# del any symbols we don't want exposed
# pylint: disable=wrong-import-position

'''
Configuration & globally-relevant values
'''

# Really just a bogus name for cases when OpenAPI is being emulated
# OpenAI API requires the model be specified, but many compaitble APIs
# have a model predetermined by the host
HOST_DEFAULT_MODEL = HOST_DEFAULT = 'HOST-DEFAULT'
OPENAI_KEY_DUMMY = 'OPENAI_DUMMY'


class attr_dict(dict):
    # XXX: Should unknown attr access return None rather than raise?
    # If so, can just do: __getattr__ = dict.get
    # __getattr__ = dict.__getitem__
    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError:
            # Substitute with more normally expected exception
            raise AttributeError(attr)
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
#
# SPDX-License-Identifier: Apache-2.0
# ogbujipt.prompting.basic

'''
Routines for building prompts for LLM query

Can be used instead of, e.g. some of the more basic langchain chains,
but designed to be much more transparent

Also, trying to avoid more complex templating such as Jinja or Mako
'''

from enum import Enum


class pdelim(Enum):
    '''
    Types of delimiters used in prompts. Prompt delimiters can be anything
    from simple spacing/formatting aids to control text such as the
    '### Instruction:\n' but in Alpaca-style prompting, for example
    '''
    FIXED_PREAMBLE = 1  # Preamble that's always set
    PRE_PREAMBLE = 2  # Set as prefix to a preamble (e.g. ### System:)
    POST_PREAMBLE = 3  # Set as prefix to a preamble (e.g. ### System:)
    INTERCONTEXT = 4  # Between multiple context sections
    PREQUERY = 5  # e.g. '### Instruction:\n' in Alpaca
    POSTQUERY = 6  # e.g. '### Response:\n' in Alpaca
    PRECONTEXT = 7  # e.g. '### Input:\n' in Alpaca instruct variation
    POSTCONTEXT = 8
    PRE_ALL_CONTEXT = 9
    POST_ALL_CONTEXT = 10
    PRE_CONTEXT_METADATA = 11   #Mostly for Airoboros (BEGINCONTEXT [..] ENDCONTEXT) format
    POST_CONTEXT_METADATA = 12  #"       "
    META_ORDERING = 100


class ordering(Enum):
    '''
    Ordering of complex prompt compoents. Of course preamble always comes 1st
    '''
    QUERY_CONTEXT = 1
    CONTEXT_QUERY = 2


def format(query, preamble='', contexts=None, delimiters=None, context_with_metadata=False):
    '''
    Build a full LLM prompt out of the actual human/user query, an optional
    preamble (e.g. system message to condition the LLM's responses),
    zero or more context items (either simple text or tuples of text/metadata
    dict), and optional delimiters

    When context_with_metadata is True, this will assume contexts is a list of 2-item tuple of strings, where
    the first item is the metadata for the context (a string) and the second item is the context string itself

    >>> from ogbujipt.prompting.basic import format
    >>> format('How are you?', preamble='You are a friendly AI who loves conversation') 
    'You are a friendly AI who loves conversation\n\nHow are you?\n'
    '''
    contexts = contexts or []
    if isinstance(contexts, str):
        contexts = [contexts]
    delimiters = delimiters or {}

    parts = []
    fixed = delimiters.get(pdelim.FIXED_PREAMBLE)
    if fixed:
        parts.append(fixed)
    if preamble:
        parts.extend([delimiters.get(pdelim.PRE_PREAMBLE, ''), preamble, '\n'])

    # XXX Maybe find a more efficient way than closures (re: function-call overhead)
    def add_context():
        '''
        Append the context info to the output parts
        '''
        if contexts:
            parts.append(delimiters.get(pdelim.PRE_ALL_CONTEXT, ''))
            for c in contexts:
                # Some prompt conventions might use a pre & post context convention
                # Some use internal delimiters only
                parts.append(delimiters.get(pdelim.PRECONTEXT, ''))
                if context_with_metadata:
                    metadata, val = c
                    parts.append(delimiters.get(pdelim.PRE_CONTEXT_METADATA, '') + '\n')
                    parts.append(metadata + '\n')
                    parts.append(delimiters.get(pdelim.POST_CONTEXT_METADATA, '') + '\n')
                if context_with_metadata:
                    metadata, val = c
                    parts.append(str(val))
                else:
                    parts.append(str(c))
                parts.append(delimiters.get(pdelim.POSTCONTEXT, ''))
                parts.append(delimiters.get(pdelim.INTERCONTEXT, '\n'))
            del parts[-1]  # Final intercontext (might be empty anyway) not needed
            parts.append(delimiters.get(pdelim.POST_ALL_CONTEXT, ''))

    def add_query():
        '''
        Append the main query to the output parts
        '''
        parts.append(delimiters.get(pdelim.PREQUERY, ''))
        parts.append(query)

    # 
    if delimiters.get(pdelim.META_ORDERING, ordering.CONTEXT_QUERY) \
            == ordering.CONTEXT_QUERY:
        add_context()
        add_query()
    else:
        add_query()
        add_context()

    # XXX: Feels a bit weird that the post-query bit must be outside the query
    # closure. Maybe needs a rename?
    parts.append(delimiters.get(pdelim.POSTQUERY, '\n'))
    full_context = ''.join(parts)
    return full_context


# Support former name, for now
context_build = format



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# test/test_config.py
from ogbujipt import config

def test_attr_dict():
    d = config.attr_dict({'a': 1, 'b': 2})
    assert d.__class__.__bases__ == (dict,)
    assert d.a == 1
    assert d.b == 2
    assert d['a'] == 1
    assert d['b'] == 2


if __name__ == '__main__':
    raise SystemExit("Attention! Run with pytest")



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# test/test_ogbujipt.py
'''
pytest test

or

pytest test/test_ogbujipt.py
'''
# ruff: noqa: E501

# import pytest

from ogbujipt.llm_wrapper import openai_chat_api

def test_oapi_first_choice_text(OPENAI_TEXT_RESPONSE_OBJECT):
    text1 = openai_chat_api.first_choice_text(OPENAI_TEXT_RESPONSE_OBJECT)
    assert text1 == '…is an exceptional employee who has made significant contributions to our company.'

def test_oapi_first_choice_message(OPENAI_MSG_RESPONSE_OBJECT):
    msg1 = openai_chat_api.first_choice_message(OPENAI_MSG_RESPONSE_OBJECT)
    assert msg1 == '…is an exceptional employee who has made significant contributions to our company.'


if __name__ == '__main__':
    raise SystemExit("Attention! Run with pytest")



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# demo/function_calling.py
'''
Demonstrate the use of OpenAI-style function calling with OgbujiPT

python demo/function_calling.py

Requires `OPENAI_API_KEY` in the environment

Hard-codes the function spec, but you can generate it from a PyDantic schema

```py
from typing import List
from pydantic import BaseModel

# PyDantic schema from which we'll generate function spec
class ExecuteStepByStepPlan(BaseModel):
    title: str
    steps: List[str]

# Generate the function spec
FUNC_SPEC = ExecuteStepByStepPlan.schema()
```

What about non-OpenAI LLM hosts? There is ongoing work in several areas.
It requires properly fine-tuned models, the right systems prompts and also suppot by the host code
Useful discussion re llama-cpp-python: https://github.com/abetlen/llama-cpp-python/discussions/397
'''

from ogbujipt.llm_wrapper import openai_chat_api, prompt_to_chat

FUNC_SPEC = {
    'title': 'ExecuteStepByStepPlan',
    'type': 'object',
    'properties': {
        'headline': {'headline': 'Headline', 'type': 'string'},
        'steps': {'title': 'Steps', 'type': 'array', 'items': {'type': 'string'}},
    },
    'required': ['headline', 'steps'],
}

# Requires OPENAI_API_KEY in environment
llm_api = openai_chat_api(model='gpt-3.5-turbo')

messages = prompt_to_chat('Explain how to poach an egg, step by step')

functions=[
        {
          'name': 'handle_steps_from_user_query',
          'description': 'Respond to a user query by specifying a series of steps',
          'parameters': FUNC_SPEC
        }
]

function_call={'name': 'handle_steps_from_user_query'}

resp = llm_api(messages=messages, functions=functions, function_call=function_call)
fc = resp.choices[0].message.function_call

if fc:
    print('Function to be called: ' + fc.name)
    print('Function call arguments: ' + fc.arguments)
else:
    print('No function call issued')



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
#
# SPDX-License-Identifier: Apache-2.0
# ogbujipt.embedding
'''
Vector DBs are useful when you have a lot of context to use with LLMs,
e.g. a large document or collection of docs. One common pattern is to create
vector indices on this text. Given an LLM prompt, the vector DB is first queried
to find the most relevant "top k" sections of text based on the prompt,
which are added as context in the ultimate LLM invocation.

For sample code see demo/chat_pdf_streamlit_ui.py

You need an LLM to turn the text into vectors for such indexing, and these
vectors are called the embeddings. You can usually create useful embeddings
with a less powerful (and more efficient) LLM.

This package provides utilities to set up a vector DB, and use it to index
chunks of text using a provided LLM model to create the embeddings.

Other common use-cases for vector DBs in LLM applications:

* Long-term LLM Memory for chat: index the entire chat history and retrieve
the most relevant and recent N messages based on the user's new message,
to give the LLM a chance to make its responses coherent and on-topic

* Cache previous LLM interactions, saving resources by retrieving previous
responses to similar questions without having to use the most powerful LLM
'''



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# ogbujipt/demo/chat_pdf_streamlit_ui.py
'''
Advanced. Retrieval Augmented Generation (RAG) AKA "Chat my PDF" demo

Use a PDF document as a knowledge base to provide context for natural language Q&A

UI: Streamlit - streamlit.io
Vector store: Qdrant
PDF to text: PyPDF2
Text to vector (embedding) model: HuggingFace all-MiniLM-L12-v2

Single-PDF support, for now, to keep the demo code simple. Can easily extend to
e.g. work with multiple docs dropped in a directory

Prerequisites. From OgbujiPT cloned dir:.

```sh
pip install streamlit watchdog PyPDF2 PyCryptodome sentence_transformers qdrant-client tiktoken
```

You'll need environment variables. See demo/.env for example settings. Load them in your environment as you see fit:
https://github.com/OoriData/OgbujiPT/discussions/36

Run the demo:

```sh
streamlit run demo/chat_pdf_streamlit_ui.py
```
'''
import os

import streamlit as st
from PyPDF2 import PdfReader

from ogbujipt.llm_wrapper import openai_chat_api, prompt_to_chat
from ogbujipt.text_helper import text_splitter
from ogbujipt.embedding.qdrant import collection

from sentence_transformers import SentenceTransformer

# Avoid re-entrace complaints from huggingface/tokenizers
os.environ['TOKENIZERS_PARALLELISM'] = 'false'

# User can set a variety of likely values to trigger use of OpenAI full-service
OPENAI = os.getenv('OPENAI', 'False') in \
    ['True', 'true', 'TRUE', 'Yes', 'yes', 'YES']
OPENAI_API_KEY = os.getenv('OPENAI_API_KEY')
LLM = os.getenv('LLM', 'LLM')  # TODO: get this from non-openai openai api hosts
LLM_BASE = os.getenv('LLM_BASE', 'my-llm-host')
LLM_TEMP = float(os.getenv('LLM_TEMP', '1'))  # LLM temperature (randomness)
N_CTX = int(os.getenv('N_CTX', '2048'))  # LLM max context size
K = int(os.getenv('K', '3'))  # how many chunks to return for query context
EMBED_CHUNK_SIZE = int(os.getenv('EMBED_CHUNK_SIZE', '500'))  # Character count used in slicing up the document
EMBED_CHUNK_OVERLAP = int(os.getenv('EMBED_CHUNK_OVERLAP', '100'))  # Character count overlap between chunks
# LLM for vector DB embeddings; will be d/led: https://huggingface.co/sentence-transformers/all-MiniLM-L12-v2
DOC_EMBEDDINGS_LLM = os.getenv('DOC_EMBEDDINGS_LLM', 'all-MiniLM-L12-v2')

CONSOLE_WIDTH = 80

PDF_USER_QUERY_PROMPT = 'Ask a question about your PDF:'

favicon = 'data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAABhGlDQ1BJQ0MgcHJvZmlsZQAAKJF9kT1Iw0AcxV9TpUUqgu0g4pChOrWLijjWKhShQqkVWnUwufQLmjQkKS6OgmvBwY/FqoOLs64OroIg+AHi6uKk6CIl/i8ptIjx4Lgf7+497t4BQqvGVLMvAaiaZWRTSTFfWBUDrwhiGGFEEZOYqc9lMml4jq97+Ph6F+dZ3uf+HINK0WSATyROMN2wiDeIZzYtnfM+cYRVJIX4nDhm0AWJH7kuu/zGueywwDMjRi47TxwhFss9LPcwqxgq8TRxVFE1yhfyLiuctzirtQbr3JO/MFTUVpa5TnMMKSxiCRmIkNFAFTVYiNOqkWIiS/tJD/+o48+QSyZXFYwcC6hDheT4wf/gd7dmaWrSTQolgf4X2/4YBwK7QLtp29/Htt0+AfzPwJXW9ddbwOwn6c2uFj0ChraBi+uuJu8BlzvAyJMuGZIj+WkKpRLwfkbfVADCt8DAmttbZx+nD0COukrfAAeHwESZstc93h3s7e3fM53+fgDTbXLN1L6DRgAAAAZiS0dEAP8A/wD/oL2nkwAAAAlwSFlzAAAuIwAALiMBeKU/dgAAAAd0SU1FB+cIEhMmDa76t/8AAAAZdEVYdENvbW1lbnQAQ3JlYXRlZCB3aXRoIEdJTVBXgQ4XAAABlElEQVRYw82XMU4CQRRA32CMQEIt26yNVBSEA2BChY0UNN4AjkE4BhyDSqmQrQkhkQqNsM1CTYKFCVg4JiuKYf9njVPOzsx7+bMzfz4oWqtYqLeKhbpmjRMNHGgDN2UnG/QXy+GfCYTga+ANqEkljBJ+bbvvgDTQaI7GndgisAtvjsZef7H0y07WA24lkTAa+GpSdQEy+a7fKhZKkkgYJXxgh1xJJYwSfmGHzaUSRgn/XLwulTAaeCbfbQCsJtW2VCKhgc9eHpNPfpCyIh37bbCaVN3maOzZY7oG2vtuTHMg/AJ4CMOf57PkxqR6H4tsKpeu8yqJhJHB/eTGnPWAkp3qSSWMEu7Z6SWphIkCn/pBChL3IXj4KhZJGA08556vAab+Mi2VSBBD20bNhsfcAthUclG2IOpP+IMEUrj4GO5IIIWrLqKQBFL43lwQTQKkcEky+iahTUaSdPxFIrZ0HEEivgfJgRLxPskOkNjG/ij9TcL2i+D/ojAxkmSzI4EULq4N+4vlsOxkA6AGnErh6naM8vwdcuxC5pdn2pAAAAAASUVORK5CYII='  # noqa E501

st.set_page_config(  # Set up Streamlit page
    page_title='Ask your PDF',
    page_icon=favicon,
    layout='centered',
    initial_sidebar_state='expanded',
    )


@st.cache_data
def load_throbber():
    # oori_throbber.gif as a data URL
    # TODO: use a less archaic format than GIF; perhaps some sort of animatable vector image
    return 'data:image/gif;base64, '  # noqa E501


@st.cache_resource
def load_embedding_model(embedding_model_name):
    # LLM will be downloaded from HuggingFace automatically
    return SentenceTransformer(embedding_model_name)


def prep_pdf():
    '''
    Oori — Ask your PDF 📄💬
    '''
    'Convert pdf content into chunks according to chunk size & overlap'
    placeholder = st.empty()

    # Load PDF & collect its text & split it into chunks
    pdf = st.session_state['pdf']
    if not pdf:
        return

    with placeholder.container():
        # Get the embedding model
        if not st.session_state['embedding_model']:
            st.session_state['embedding_model'] = load_embedding_model(embedding_model_name=DOC_EMBEDDINGS_LLM)
        emb_model = st.session_state['embedding_model']

        # Vectorizes chunks for sLLM lookup
        # XXX: Look up rules around uploaded object names
        kb = collection(pdf.name, emb_model)  # in-memory vector DB instance

        # Show throbber, embed the PDF, and get ready for similarity search
        embedding_placeholder = st.container()
        embedding_placeholder.write('Embedding PDF...')

        # Load throbber from cache
        throbber = load_throbber()
        embedding_placeholder.image(throbber)

        # Prepare a vector knowledgebase based on the pdf contents
        # Use st.session_state to avoid unnecessary reprocessing/reloading
        pdf_reader = PdfReader(pdf)
        text = ''.join((page.extract_text() for page in pdf_reader.pages))
        chunks = text_splitter(
            text, 
            chunk_size=EMBED_CHUNK_SIZE,
            chunk_overlap=EMBED_CHUNK_OVERLAP,
            separator='\n')

        # Update vector DB collection, insert the text chunks & update app state
        kb.update(texts=chunks)
        st.session_state['kb'] = kb  # Update state
    placeholder.empty()

    # Get the user query
    st.text_input(
        label=PDF_USER_QUERY_PROMPT,
        key='user_query_str',
        on_change=query_llm,
        args=(st.session_state['openai_api'],))


def query_llm(openai_api):
    '''
    Oori — Ask your PDF 📄💬
    '''
    kb = st.session_state['kb']
    user_query = st.session_state['user_query_str']

    # Placeholder for throbber & LLM response
    response_placeholder = st.empty()
    # Load throbber from cache
    throbber = load_throbber()
    response_placeholder.image(throbber)

    docs = kb.search(user_query, limit=K)

    # Concatenate text chunks for insertion into prompt
    gathered_chunks = '\n\n'.join(doc.payload['_text'] for doc in docs if doc.payload)

    # In examples I've seen rag-stuffed content seems to go in the system message
    # https://www.pinecone.io/learn/context-aware-chatbot-with-vercel-ai-sdk/#Step-3.-Adding-Context

    # In practice we'd use word loom to load the propts, as demoed in multiprocess.py
    sys_prompt = '''\
You are a helpful assistant, who answers questions directly and as briefly as possible.
Consider the following context and answer the user\'s question.
If you cannot answer with the given context, just say so.\n\n'''
    sys_prompt += gathered_chunks + '\n\n'
    messages = prompt_to_chat(user_query, system=sys_prompt)

    print('  MESSAGES FOR LLM:  '.center(CONSOLE_WIDTH, '='), '\n', messages)

    # Remember max Token length default is 16)
    oapi = st.session_state['openai_api']
    # Need to remove old system messages for subsequent queries
    st.session_state['messages'].extend(messages)
    # print(st.session_state['messages'], '\n', '-'*10)
    response = oapi(st.session_state['messages'], temperature=LLM_TEMP, max_tokens=1024)

    print('\nFull response data from LLM:\n', response)

    # Response is a json-like object; extract the text
    response_text = oapi.first_choice_message(response)
    print('\nResponse text from LLM:\n', response_text)

    st.session_state['messages'].append({'role': 'assistant', 'content': response})
    response_placeholder.write(response_text)
    # print(st.session_state['messages'], '\n', '-'*10)


def main():
    '''
    Oori — Ask your PDF 📄💬
    '''
    # Streamlit treats function docstrings as magic strings for user display. Use comments instead
    # Set up LLM host connection & launch the main loop
    # Use OpenAI API if specified, otherwise emulate with supplied host, etc. for self-hosted LLM
    if OPENAI:
        assert not (LLM_BASE), 'Don\'t use LLM_BASE with the OPENAI variable'
        oapi = openai_chat_api(api_key=OPENAI_API_KEY, model=(LLM or 'gpt-3.5-turbo'))
    else:
        oapi = openai_chat_api(base_url=LLM_BASE, model=LLM)

    st.session_state['embedding_model'] = None
    st.session_state['openai_api'] = oapi
    # Chat message history in memory
    st.session_state['messages'] = []

    # Create file upload box on Streamlit, set from the user's upload
    # Use st.session_state to avoid unnessisary reprocessing/reloading
    st.file_uploader('Upload a PDF', type=['pdf'], accept_multiple_files=False,
                     on_change=prep_pdf, key='pdf')


main()  # Code to execute



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# ogbujipt.model_styles.async_helper

'''
Coroutines to make it a little easier to multitask LLM access
using Python asyncio
'''
import sys
import asyncio

async def console_progress_indicator(pause=0.5, file=sys.stderr):
    '''
    Simple progress indicator for the console. Just prints dots.

    pause - seconds between each dot printed to console, default half a sec

    file - file for dots output, default STDERR
    '''
    while True:
        print('.', end='', flush=True, file=file)
        await asyncio.sleep(pause)



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# ogbujipt.embedding.pgvector_message

'''
Vector embeddings DB feature for messaging (chat, etc.) using PGVector
'''

from uuid import UUID
from datetime import datetime, timezone
from typing import Iterable
import warnings

from ogbujipt.config import attr_dict
from ogbujipt.embedding.pgvector import PGVectorHelper, asyncpg, process_search_response

__all__ = ['MessageDB']

# ------ SQL queries ---------------------------------------------------------------------------------------------------
# PG only supports proper query arguments (e.g. $1, $2, etc.) for values, not for table or column names
# Table names are checked to be legit sequel table names, and embed_dimension is assured to be an integer

CREATE_MESSAGE_TABLE = '''-- Create a table to hold individual messages (e.g. from a chatlog) and their metadata
CREATE TABLE IF NOT EXISTS {table_name} (
    ts TIMESTAMP WITH TIME ZONE,              -- timestamp of the message
    history_key UUID,                         -- uunique identifier for contextual message history
    role TEXT,                                -- role of the message (meta ID such as 'system' or user,
                                              -- or an ID associated with the sender)
    content TEXT NOT NULL,                    -- text content of the message
    embedding VECTOR({embed_dimension}),      -- embedding vectors (array dimension)
    metadata JSON                             -- additional metadata of the message
);
'''

INSERT_MESSAGE = '''-- Insert a message into a chatlog
INSERT INTO {table_name} (
    history_key,
    role,
    content,
    embedding,
    ts,
    metadata
) VALUES ($1, $2, $3, $4, $5, $6);
'''

CLEAR_MESSAGE = '''-- Deletes matching messages
DELETE FROM {table_name}
WHERE
    history_key = $1
'''

RETURN_MESSAGES_BY_HISTORY_KEY = '''-- Get entire chatlog by history key
SELECT
    ts,
    role,
    content,
    metadata
FROM
    {table_name}
WHERE
    history_key = $1
{since_clause}
ORDER BY
    ts DESC
{limit_clause};
'''

SEMANTIC_QUERY_MESSAGE_TABLE = '''-- Find messages with closest semantic similarity
SELECT
    1 - (embedding <=> $1) AS cosine_similarity,
    ts,
    role,
    content,
    metadata
FROM
    {table_name}
WHERE
    history_key = $2
ORDER BY
    cosine_similarity DESC
LIMIT $3;
'''
# ------ SQL queries ---------------------------------------------------------------------------------------------------

class MessageDB(PGVectorHelper):
    ''' Specialize PGvectorHelper for messages, e.g. chatlogs '''
    async def create_table(self):
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await conn.execute(
                    CREATE_MESSAGE_TABLE.format(
                        table_name=self.table_name,
                        embed_dimension=self._embed_dimension
                    )
                )

    async def insert(
            self,
            history_key: UUID,
            role: str,
            content: str,
            # Timestamp later in order than SQL might suggest because of the defaulting
            timestamp: datetime | None = None,
            metadata: dict | None = None
    ) -> None:
        '''
        Update a table with one message's embedding

        Args:
            history_key (str): history key (unique identifier) of the chatlog this message belongs to

            role (str): role of the message (e.g. system, user, assistant, tool)

            content (str): text content of the message

            timestamp (datetime, optional): timestamp of the message, defaults to current time

            metadata (dict[str, str], optional): additional metadata of the message

        Returns:
            generator which yields the rows os the query results ass attributable dicts
        '''
        if not timestamp:
            timestamp = datetime.utcnow().replace(tzinfo=timezone.utc)

        # Get the embedding of the content as a PGvector compatible list
        content_embedding = self._embedding_model.encode(content)

        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await conn.execute(
                    INSERT_MESSAGE.format(table_name=self.table_name),
                    history_key,
                    role,
                    content,
                    content_embedding.tolist(),
                    timestamp,
                    metadata
                )   

    async def insert_many(
            self,
            content_list: Iterable[tuple[str, list[str]]]
    ) -> None:
        '''
        Update table with one or (presumably) more message embedding

        Semantically equivalent to multiple insert calls, but uses executemany for efficiency

        Args:
            content_list: List of tuples, each of the form: (history_key, role, text, timestamp, metadata)
        '''
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await conn.executemany(
                    INSERT_MESSAGE.format(table_name=self.table_name),
                    (
                        (hk, role, text, self._embedding_model.encode(text), ts, metadata)
                        for hk, role, text, ts, metadata in content_list
                    )
                )

    async def clear(
            self,
            history_key: UUID
    ) -> None:
        '''
        Remove all entries in a message history entries

        Args:
            history_key (str): history key (unique identifier) to match
        '''
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await conn.execute(
                    CLEAR_MESSAGE.format(
                        table_name=self.table_name,
                    ),
                    history_key
                )

    # XXX: Change to a generator
    async def get_messages(
            self,
            history_key: UUID | str,
            since: datetime | None = None,
            limit: int = 0
    ): # -> list[asyncpg.Record]:
        '''
        Retrieve entries in a message history

        Args:
            history_key (str): history key (unique identifier) to match; string or object
            since (datetime, optional): only return messages after this timestamp
            limit (int, optional): maximum number of messages to return. Default is all messages
        Returns:
            generates asyncpg.Record instances of resulting messages
        '''
        qparams = [history_key]
        async with self.pool.acquire() as conn:
            if not isinstance(history_key, UUID):
                history_key = UUID(history_key)
                # msg = f'history_key must be a UUID, not {type(history_key)} ({history_key}))'
                # raise TypeError(msg)
            if not isinstance(since, datetime) and since is not None:
                msg = 'since must be a datetime or None'
                raise TypeError(msg)
            if since:
                # Really don't need the ${len(qparams) + 1} trick here, but used for consistency
                since_clause = f' AND ts > ${len(qparams) + 1}'
                qparams.append(since)
            else:
                since_clause = ''
            if limit:
                limit_clause = f'LIMIT ${len(qparams) + 1}'
                qparams.append(limit)
            else:
                limit_clause = ''
            message_records = await conn.fetch(
                RETURN_MESSAGES_BY_HISTORY_KEY.format(
                    table_name=self.table_name,
                    since_clause=since_clause,
                    limit_clause=limit_clause
                ),
                *qparams
            )

        return (attr_dict({
                'ts': record['ts'],
                'role': record['role'],
                'content': record['content'],
                'metadata': record['metadata']
            }) for record in message_records)

        # async for record in message_records:
        #     yield attr_dict({
        #         'ts': record['ts'],
        #         'role': record['role'],
        #         'content': record['content'],
        #         'metadata': record['metadata']
        #     })
    
    async def get_table(self, history_key):
        # Deprecated
        warnings.warn('pgvector_message.get_table() is deprecated. Use get_messages().', DeprecationWarning)
        return list(await self.get_messages(history_key))

    async def search(
            self,
            history_key: UUID,
            text: str,
            since: datetime | None = None,
            limit: int = 1
    ) -> list[asyncpg.Record]:
        '''
        Similarity search documents using a query string

        Args:
            history_key (str): history key for the conversation to query
            text (str): string to compare against items in the table
            since (datetime, optional): only return results after this timestamp
            limit (int, optional): maximum number of messages to return; for top-k type query. Default is 1
        Returns:
            list[asyncpg.Record]: list of search results
                (asyncpg.Record objects are similar to dicts, but allow for attribute-style access)
        '''
        if not isinstance(limit, int):
            raise TypeError('limit must be an integer')

        # Get the embedding of the query string as a PGvector compatible list
        query_embedding = list(self._embedding_model.encode(text))

        # Search the table
        async with self.pool.acquire() as conn:
            records = await conn.fetch(
                SEMANTIC_QUERY_MESSAGE_TABLE.format(
                    table_name=self.table_name
                ),
                query_embedding,
                history_key,
                limit
            )

        search_results = [
            {
                'ts': record['ts'],
                'role': record['role'],
                'content': record['content'],
                'metadata': record['metadata'],
                'cosine_similarity': record['cosine_similarity']
            }
            for record in records
        ]

        return process_search_response(search_results)



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# test/test_model_style.py
# FIXME: For consistency let's migrate this to pytest
from unittest.mock import patch, Mock

from ogbujipt.prompting import model_style
from ogbujipt.llm_wrapper import openai_api

import unittest

# Should no longer be necessary, with mocking
#     try:
#         import httpx  # noqa
#     except ImportError:
#         raise RuntimeError('Needs httpx installed. Try pip install httpx')


class Test_hosted_model_openAI(unittest.TestCase):

    @patch('httpx.get')
    @patch('openai.base_url', 'https://api.openai.com/v1')  # Use your actual API URL
    def test_hosted_model_openai(self, mock_get):
        mock_response = Mock()
        mock_response.json.return_value = {
            'data': [{'id': 'model_1'}, {'id': 'model_2'}]
        }
        mock_get.return_value = mock_response
        
        llm = openai_api()
        result = llm.hosted_model()

        self.assertEqual(result, 'model_1')
        
        result = llm.available_models()

        self.assertEqual(result, ['model_1', 'model_2'])

    def test_model_style_from_name(self):

        result1 = model_style.model_style_from_name('path/wizardlm-13b-v1.0-uncensored.ggmlv3.q6_K.bin')
        self.assertEqual(str(result1), "[<style.WIZARD: 4>]")


if __name__ == '__main__':
    raise SystemExit("Attention! Run with pytest")



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# ogbujipt.word_loom

'''
Routines to help with processing text in the word loom convention:
https://github.com/OoriData/OgbujiPT/wiki/Word-Loom:-A-format-for-managing-language-for-AI-LLMs-(including-prompts)
'''

import io
import tomli


class text_item(str):
    '''
    Text or template for use with LLM tools
    Largely keeps metadata around language, template markers, etc.

    >>> from ogbujipt.word_loom import T
    >>> t = T('spam', lang='en)
    'spam'
    >>> t.lang
    'en'
    '''

    def __new__(cls, value, lang, markers=None):
        assert isinstance(value, str)
        self = super(text_item, cls).__new__(cls, value)
        self.lang = lang
        self.markers = markers
        return self

    def __repr__(self):
        return u'T(' + str(self) + ')'


T = text_item


# XXX Defaulting to en leaves a bit too imperialist a flavor, really
def load(fp_or_str, lang='en'):
    '''
    Read a word loom and return the tables as top-level result mapping
    Loads the TOML, then selects text by given language

    fp_or_str - file-like object or string containing TOML
    lang - select oly texts in this language (default: 'en')

    >>> from ogbujipt import word_loom
    >>> with open('myprompts.toml', mode='rb') as fp:
    >>>     loom = word_loom.load(fp)
    '''
    # Ensure we have a file-like object
    if isinstance(fp_or_str, str):
        fp_or_str = io.BytesIO(fp_or_str.encode('utf-8'))
    elif isinstance(fp_or_str, bytes):
        fp_or_str = io.BytesIO(fp_or_str)
    # Load TOML
    loom_raw = tomli.load(fp_or_str)
    # Select text by language
    # FIXME: Only top level, for now. Presumably we'll want to support scoping
    texts = {}
    default_lang = loom_raw.get('lang', None)
    for k, v in loom_raw.items():
        if not isinstance(v, dict) or 'text' not in v:
            # Skip top-level items
            continue
        if v.get('lang') == lang or ('lang' not in v and lang == default_lang):
            texts[k] = T(v['text'], lang, markers=v.get('markers'))
    return texts



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# ogbujipt.embedding.pgvector_data_doc

'''
Vector databases embeddings using PGVector
'''

import warnings
from typing import Iterable

from ogbujipt.embedding.pgvector import PGVectorHelper, asyncpg, process_search_response

__all__ = ['DocDB']

# ------ SQL queries ---------------------------------------------------------------------------------------------------
# PG only supports proper query arguments (e.g. $1, $2, etc.) for values, not for table or column names
# Table names are checked to be legit sequel table names, and embed_dimension is assured to be an integer

CREATE_TABLE_BASE = '''-- Create a table to hold embedded documents or data
CREATE TABLE IF NOT EXISTS {{table_name}} (
    id BIGSERIAL PRIMARY KEY,
    embedding VECTOR({{embed_dimension}}),  -- embedding vectors (array dimension)
    content TEXT NOT NULL,                  -- text content of the chunk
    tags TEXT[]                             -- tags associated with the chunk
{extra_fields});
'''

CREATE_DOC_TABLE = CREATE_TABLE_BASE.format(extra_fields='''\
,    title TEXT,                            -- title of file
    page_numbers INTEGER[]                  -- page number of the document that the chunk is found in
''')

CREATE_DATA_TABLE = CREATE_TABLE_BASE.format(extra_fields='')

INSERT_BASE = '''-- Insert a document into a table
INSERT INTO {{table_name}} (
    embedding,
    content,
    tags
    {extra_fields}) VALUES ($1, $2, $3{extra_vals});
'''

INSERT_DOCS = INSERT_BASE.format(extra_fields='''\
,    title,
    page_numbers
''', extra_vals=', $4, $5')

INSERT_DATA = INSERT_BASE.format(extra_fields='', extra_vals='')

QUERY_TABLE_BASE = '''-- Semantic search a document
SELECT * FROM
-- Subquery to calculate cosine similarity, required to use the alias in the WHERE clause
(
    SELECT
        1 - (embedding <=> $1) AS cosine_similarity,
        content,
        tags
        {extra_fields}
    FROM
        {{table_name}}
) subquery
{{where_clauses}}
ORDER BY
    cosine_similarity DESC
{{limit_clause}};
'''

QUERY_DOC_TABLE = QUERY_TABLE_BASE.format(extra_fields='''\
,        title,
        page_numbers
''')

QUERY_DATA_TABLE = QUERY_TABLE_BASE.format(extra_fields='')

TITLE_WHERE_CLAUSE = 'title = {query_title}  -- Equals operator\n'

PAGE_NUMBERS_WHERE_CLAUSE = 'page_numbers && {query_page_numbers}  -- Overlap operator \n'

TAGS_WHERE_CLAUSE_CONJ = 'tags @> {tags}  -- Contains operator\n'
TAGS_WHERE_CLAUSE_DISJ = 'tags && {tags}  -- Overlap operator\n'

THRESHOLD_WHERE_CLAUSE = 'cosine_similarity >= {query_threshold}\n'
# ------ SQL queries ---------------------------------------------------------------------------------------------------


# XXX: Data vs doc DB can probably be modularized further, but this will do for now
class DataDB(PGVectorHelper):
    ''' Specialize PGvectorHelper for data (snippets) '''
    async def create_table(self) -> None:
        '''
        Create the table to hold embedded documents
        '''
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await conn.execute(
                    CREATE_DATA_TABLE.format(
                        table_name=self.table_name,
                        embed_dimension=self._embed_dimension)
                    )
    
    async def insert(
            self,
            content: str,
            tags: list[str] = []
    ) -> None:
        '''
        Update a table with one embedded document

        Args:
            content (str): text content of the document

            tags (list[str], optional): tags associated with the document
        '''
        # Get the embedding of the content as a PGvector compatible list
        content_embedding = self._embedding_model.encode(content)

        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await conn.execute(
                    INSERT_DATA.format(table_name=self.table_name),
                    content_embedding,
                    content,
                    tags
                )

    async def insert_many(
            self,
            content_list: Iterable[tuple[str, list[str]]]
    ) -> None:
        '''
        Update a table with one or (presumably) more embedded documents

        Semantically equivalent to multiple insert calls, but uses executemany for efficiency

        Args:
            content_list: List of tuples, each of the form: (content, tags)
        '''
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await conn.executemany(
                    INSERT_DATA.format(table_name=self.table_name),
                    (
                        # Does this need to be .tolist()?
                        (self._embedding_model.encode(content), content, tags)
                        for content, tags in content_list
                    )
                )

    async def search(
            self,
            text: str,
            tags: list[str] | None = None,
            threshold: float | None = None,
            limit: int = 0,
            conjunctive: bool = True,
            query_tags: list[str] | None = None,
    ) -> list[asyncpg.Record]:
        '''
        Similarity search documents using a query string

        Args:
            text (str): string to compare against items in the table.
                This will be a vector/fuzzy/nearest-neighbor type search.

            tags (list[str], optional): tags associated with the document to compare against items in the table.
                Each individual tag must match exactly, but see the conjunctive param
                for how multiple tags are interpreted.

            limit (int, optional): maximum number of results to return (useful for top-k query)
                Default is no limit

            conjunctive (bool, optional): whether to use conjunctive (AND) or disjunctive (OR) matching
                in the case of multiple tags. Defaults to True.

        Returns:
            generator which yields the rows os the query results ass attributable dicts
        '''
        if query_tags is not None:
            warnings.warn('query_tags is deprecated. Use tags instead.', DeprecationWarning)
            tags = query_tags
        # else:
        #     if not isinstance(query_tags, list):
        #         raise TypeError('query_tags must be a list of strings')
        #     if not all(isinstance(tag, str) for tag in query_tags):
        #         raise TypeError('query_tags must be a list of strings')
        if threshold is not None:
            if not isinstance(threshold, float):
                raise TypeError('threshold must be a float')
            if (threshold < 0) or (threshold > 1):
                raise ValueError('threshold must be between 0 and 1')

        if not isinstance(limit, int):
            raise TypeError('limit must be an integer')  # Guard against injection

        # Get the embedding of the query string as a PGvector compatible list
        query_embedding = list(self._embedding_model.encode(text))

        tags_where_clause = TAGS_WHERE_CLAUSE_CONJ if conjunctive else TAGS_WHERE_CLAUSE_DISJ

        # Build where clauses
        if (tags is None) and (threshold is None):
            # No where clauses, so don't bother with the WHERE keyword
            where_clauses = ''
            query_args = [query_embedding]
        else:  # construct where clauses
            param_count = 1
            clauses = []
            query_args = [query_embedding]
            if tags is not None:
                param_count += 1
                query_args.append(tags)
                clauses.append(tags_where_clause.format(tags=f'${param_count}'))
            if threshold is not None:
                param_count += 1
                query_args.append(threshold)
                clauses.append(THRESHOLD_WHERE_CLAUSE.format(query_threshold=f'${param_count}'))
            clauses = 'AND\n'.join(clauses)  # TODO: move this into the fstring below after py3.12
            where_clauses = f'WHERE\n{clauses}'

        if limit:
            limit_clause = f'LIMIT {limit}\n'
        else:
            limit_clause = ''

        # Execute the search via SQL
        async with self.pool.acquire() as conn:
            search_results = await conn.fetch(
                QUERY_DATA_TABLE.format(
                    table_name=self.table_name,
                    where_clauses=where_clauses,
                    limit_clause=limit_clause,
                ),
                *query_args
            )
        return process_search_response(search_results)


class DocDB(PGVectorHelper):
    ''' Specialize PGvectorHelper for documents '''
    async def create_table(self) -> None:
        '''
        Create the table to hold embedded documents
        '''
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await conn.execute(
                    CREATE_DOC_TABLE.format(
                        table_name=self.table_name,
                        embed_dimension=self._embed_dimension)
                    )
    
    async def insert(
            self,
            content: str,
            title: str = 'NULL',
            page_numbers: list[int] = [],
            tags: list[str] = []
    ) -> None:
        '''
        Update a table with one embedded document

        Args:
            content (str): text content of the document

            title (str, optional): title of the document

            page_numbers (list[int], optional): page number of the document that the chunk is found in

            tags (list[str], optional): tags associated with the document
        '''
        # Get the embedding of the content as a PGvector compatible list
        content_embedding = self._embedding_model.encode(content)

        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await conn.execute(
                    INSERT_DOCS.format(table_name=self.table_name),
                    content_embedding.tolist(),
                    content,
                    tags,
                    title,
                    page_numbers
                )

    async def insert_many(
            self,
            content_list: Iterable[tuple[str, list[str], str | None, list[int]]]
    ) -> None:
        '''
        Update a table with one or more embedded documents

        Semantically equivalent to multiple insert_doc calls, but uses executemany for efficiency

        Args:
            content_list: List of tuples, each of the form: (content, tags, title, page_numbers)
        '''
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await conn.executemany(
                    INSERT_DOCS.format(table_name=self.table_name),
                    (
                        (self._embedding_model.encode(content), content, tags, title, page_numbers)
                        for content, tags, title, page_numbers in content_list
                    )
                )

    async def search(
            self,
            text: str,
            query_title: str | None = None,
            query_page_numbers: list[int] | None = None,
            tags: list[str] | None = None,
            threshold: float | None = None,
            limit: int = 0,
            conjunctive: bool = True,
            query_tags: list[str] | None = None,
    ) -> list[asyncpg.Record]:
        '''
        Similarity search documents using a query string

        Args:
            text (str): string to compare against items in the table.
                This will be a vector/fuzzy/nearest-neighbor type search.

            query_title (str, optional): title of the document to compare against items in the table.

            query_page_numbers (list[int], optional): target page number in the document for query string comparison.

            tags (list[str], optional): tags associated with the document to compare against items in the table.
                Each individual tag must match exactly, but see the conjunctive param
                for how multiple tags are interpreted.

            limit (int, optional): maximum number of results to return (useful for top-k query)
                Default is no limit

            conjunctive (bool, optional): whether to use conjunctive (AND) or disjunctive (OR) matching
                in the case of multiple tags. Defaults to True.
        
        Returns:
            generator which yields the rows os the query results ass attributable dicts
        '''
        if query_tags is not None:
            warnings.warn('query_tags is deprecated. Use tags instead.', DeprecationWarning)
            tags = query_tags
        if threshold is not None:
            if not isinstance(threshold, float):
                raise TypeError('threshold must be a float')
            if (threshold < 0) or (threshold > 1):
                raise ValueError('threshold must be between 0 and 1')

        if not isinstance(limit, int):
            raise TypeError('limit must be an integer')  # Guard against injection

        # Get the embedding of the query string as a PGvector compatible list
        query_embedding = list(self._embedding_model.encode(text))

        tags_where_clause = TAGS_WHERE_CLAUSE_CONJ if conjunctive else TAGS_WHERE_CLAUSE_DISJ

        # Build where clauses
        if (query_title is None) and (query_page_numbers is None) and (tags is None) and (threshold is None):
            # No where clauses, so don't bother with the WHERE keyword
            where_clauses = ''
            query_args = [query_embedding]
        else:  # construct where clauses
            param_count = 1
            clauses = []
            query_args = [query_embedding]
            if query_title is not None:
                param_count += 1
                query_args.append(query_title)
                clauses.append(TITLE_WHERE_CLAUSE.format(query_title=f'${param_count}'))
            if query_page_numbers is not None:
                param_count += 1
                query_args.append(query_page_numbers)
                clauses.append(PAGE_NUMBERS_WHERE_CLAUSE.format(query_page_numbers=f'${param_count}'))
            if tags is not None:
                param_count += 1
                query_args.append(tags)
                clauses.append(tags_where_clause.format(tags=f'${param_count}'))
            if threshold is not None:
                param_count += 1
                query_args.append(threshold)
                clauses.append(THRESHOLD_WHERE_CLAUSE.format(query_threshold=f'${param_count}'))
            clauses = 'AND\n'.join(clauses)  # TODO: move this into the fstring below after py3.12
            where_clauses = f'WHERE\n{clauses}'

        if limit:
            limit_clause = f'LIMIT {limit}\n'
        else:
            limit_clause = ''

        # Execute the search via SQL
        async with self.pool.acquire() as conn:
            search_results = await conn.fetch(
                QUERY_DOC_TABLE.format(
                    table_name=self.table_name,
                    where_clauses=where_clauses,
                    limit_clause=limit_clause,
                ),
                *query_args
            )
        return process_search_response(search_results)



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# test/test_llm_wrapper.py
import os
from ogbujipt import config
from ogbujipt.llm_wrapper import openai_api, openai_chat_api, prompt_to_chat # , DUMMY_MODEL

try:
    import respx
except ImportError:
    raise RuntimeError('respx not installed. Please see wiki for setup: https://github.com/OoriData/OgbujiPT/wiki/Notes-for-contributors')  # noqa: E501

from httpx import Response

if 'OPENAI_API_KEY' in os.environ:
    del os.environ['OPENAI_API_KEY']

@respx.mock
def test_openai_llm_wrapper():
    # httpx_mock.add_response(
    #     url='http://127.0.0.1:8000/v1/models',
    #     json={'object':'list','data':[{'id':'model1','object':'model','owned_by':'me','permissions':[]}]})
    # httpx_mock.add_response(
    #     url='https://api.openai.com/v1/models',
    #     json={'object':'list','data':[{'id':'model1','object':'model','owned_by':'me','permissions':[]}]})

    respx.get('http://127.0.0.1:8000/v1/models').mock(return_value=Response(200))
    # FIXME: Figure out the right patterns for checking the HTTP requests
    # route1 = respx.get('http://127.0.0.1:8000/v1/models').mock(return_value=Response(200))

    host = 'http://127.0.0.1'
    api_key = 'jsbdflkajsdhfklajshdfkljalk'
    port = '8000'
    debug = True
    # model = DUMMY_MODEL
    rev = 'v1'
    base_url = f'{host}:{port}/{rev}'

    test_model = openai_chat_api(base_url=base_url, api_key=api_key, debug=debug)
    # assert route1.called

    assert test_model.api_key == api_key
    assert test_model.parameters.debug == debug
    assert test_model.model is None

    # Not OpenAI
    test_model = openai_api(base_url=base_url, debug=debug)

    assert test_model.api_key == config.OPENAI_KEY_DUMMY
    assert test_model.parameters.debug == debug
    assert test_model.model is None

    test_model = openai_chat_api(base_url=base_url, debug=debug)

    assert test_model.api_key == config.OPENAI_KEY_DUMMY
    assert test_model.parameters.debug == debug
    assert test_model.model is None
    assert test_model.base_url == base_url

    test_model = openai_chat_api(base_url=base_url, api_key=api_key, debug=debug)

    assert test_model.api_key == api_key
    assert test_model.parameters.debug == debug
    assert test_model.model is None


def test_prompt_to_chat():
    chatified = prompt_to_chat('test')
    assert chatified == [{'content': 'test', 'role': 'user'}]


if __name__ == '__main__':
    raise SystemExit("Attention! Run with pytest")



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# test/test_word_loom.py
'''
pytest test

or

pytest test/test_text_splitter.py
'''
# ruff: noqa: E501

import os
import sys
# import pkgutil

import pytest

from ogbujipt import word_loom

# Following only works to get the content as octets
# data = pkgutil.get_data('ogbujipt', 'resources/wordloom/sample.toml')


@pytest.fixture
def SAMPLE_TOML_STR():
    # Load language material (word loom) format
    fpath = os.path.dirname(sys.modules['ogbujipt'].__file__)

    with open(os.path.join(fpath, 'resources/wordloom/sample.toml'), 'rb') as fp:
        toml_content = fp.read()
    return toml_content


@pytest.fixture
def SAMPLE_TOML_FP():
    # Load language material (word loom) format
    fpath = os.path.dirname(sys.modules['ogbujipt'].__file__)

    return open(os.path.join(fpath, 'resources/wordloom/sample.toml'), 'rb')

def test_load_fp_vs_str(SAMPLE_TOML_STR, SAMPLE_TOML_FP):
    loom1 = word_loom.load(SAMPLE_TOML_STR)
    loom2 = word_loom.load(SAMPLE_TOML_FP)
    SAMPLE_TOML_FP.close()
    assert loom1 == loom2


def test_sample_texts_check(SAMPLE_TOML_STR):
    # print(SAMPLE_TOML)
    loom = word_loom.load(SAMPLE_TOML_STR)
    assert list(sorted(loom.keys())) == ['davinci3_instruct_system', 'i18n_context', 'write_i18n_advocacy']
    assert list(sorted([v[:20] for v in loom.values()])) == ['Internationalization', 'Obey the instruction', '{davinci3_instruct_s']
    assert [v.markers or [] for v in loom.values()] == [[], [], ['davinci3_instruct_system', 'i18n_context']]
    assert loom['davinci3_instruct_system'].lang == 'en'

    # Default language is English
    loom1 = word_loom.load(SAMPLE_TOML_STR, lang='en')
    assert loom1 == loom

    loom = word_loom.load(SAMPLE_TOML_STR, lang='fr')
    assert list(sorted(loom.keys())) == ['hardcoded_food', 'translate_request']
    assert list(sorted([v[:20] for v in loom.values()])) == ['Comment dit-on en an', 'pomme de terre']
    assert [v.markers or [] for v in loom.values()] == [['hardcoded_food'], []]
    assert loom['hardcoded_food'].lang == 'fr'


if __name__ == '__main__':
    raise SystemExit("Attention! Run with pytest")



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# test/embedding/test_pgvector.py
'''
Set up a mock Postgres instance with the following commands 
(make sure you don't have anything running on port 0.0.0.0:5432))):
docker pull ankane/pgvector
docker run --name mock-postgres -p 5432:5432 \
    -e POSTGRES_USER=mock_user -e POSTGRES_PASSWORD=mock_password -e POSTGRES_DB=mock_db \
    -d ankane/pgvector

Then run the tests with:
pytest test

or

pytest test/embedding/test_pgvector.py

Uses fixtures from ../conftest.py

'''

import pytest
# from unittest.mock import MagicMock, patch
from unittest.mock import MagicMock, DEFAULT  # noqa: F401

# from ogbujipt.embedding.pgvector import DocDB
import numpy as np

pacer_copypasta = [  # Demo document
    ('The FitnessGram™ Pacer Test is a multistage aerobic capacity test that progressively gets more difficult as it'
    ' continues.'), 
    'The 20 meter pacer test will begin in 30 seconds. Line up at the start.', 
    'The running speed starts slowly, but gets faster each minute after you hear this signal.', 
    '[beep] A single lap should be completed each time you hear this sound.', 
    '[ding] Remember to run in a straight line, and run as long as possible.', 
    'The second time you fail to complete a lap before the sound, your test is over.', 
    'The test will begin on the word start. On your mark, get ready, start.'
]

# XXX: This is to get around the fact that we can't mock the SentenceTransformer class without importing it - Kai
class SentenceTransformer(object):
    def __init__(self, model_name_or_path):
        self.encode = MagicMock()


@pytest.mark.asyncio
async def test_PGv_embed_pacer(DB):
    dummy_model = SentenceTransformer('mock_transformer')
    dummy_model.encode.return_value = np.array([1, 2, 3])
    # Insert data
    for index, text in enumerate(pacer_copypasta):   # For each line in the copypasta
        await DB.insert(                            # Insert the line into the table
            content=text,                            # The text to be embedded
            title=f'Pacer Copypasta line {index}',   # Title metadata
            page_numbers=[1, 2, 3],                  # Page number metadata
            tags=['fitness', 'pacer', 'copypasta'],  # Tag metadata
        )

    assert await DB.count_items() == len(pacer_copypasta), Exception("Not all documents inserted")

    # search table with perfect match
    search_string = '[beep] A single lap should be completed each time you hear this sound.'
    sim_search = await DB.search(text=search_string, limit=3)
    assert sim_search is not None, Exception("No results returned from perfect search")

    await DB.drop_table()

@pytest.mark.asyncio
async def test_PGv_embed_many_pacer(DB):
    dummy_model = SentenceTransformer('mock_transformer')
    dummy_model.encode.return_value = np.array([1, 2, 3])
    # Insert data using insert_many()
    documents = (
        (
            text,
            ['fitness', 'pacer', 'copypasta'],
            f'Pacer Copypasta line {index}',
            [1, 2, 3]
        )
        for index, text in enumerate(pacer_copypasta)
    )
    await DB.insert_many(documents)

    assert await DB.count_items() == len(pacer_copypasta), Exception("Not all documents inserted")

    # Search table with perfect match
    search_string = '[beep] A single lap should be completed each time you hear this sound.'
    sim_search = await DB.search(text=search_string, limit=3)
    assert sim_search is not None, Exception("No results returned from perfect search")

    await DB.drop_table()


@pytest.mark.asyncio
async def test_PGv_search_filtered(DB):
    dummy_model = SentenceTransformer('mock_transformer')
    def encode_tweaker(*args, **kwargs):
        if args[0].startswith('Text'):
            return np.array([1, 2, 3])
        else:
            return np.array([100, 300, 500])

    dummy_model.encode.side_effect = encode_tweaker
    # Need to replace the default encoder set up by the fixture
    DB._embedding_model = dummy_model
    # Insert data
    for index, text in enumerate(pacer_copypasta):   # For each line in the copypasta
        await DB.insert(                            # Insert the line into the table
            content=text,                            # The text to be embedded
            title='Pacer Copypasta',   # Title metadata
            page_numbers=[index],                    # Page number metadata
            tags=['fitness', 'pacer', 'copypasta'],  # Tag metadata
        )

    assert await DB.count_items() == len(pacer_copypasta), Exception("Not all documents inserted")

    # search table with filtered match
    search_string = '[beep] A single lap should be completed each time you hear this sound.'
    sim_search = await DB.search(
        text=search_string,
        query_title='Pacer Copypasta',
        query_page_numbers=[3],
        tags=['pacer'],
        conjunctive=False
    )
    assert sim_search is not None, Exception("No results returned from filtered search")

    #Test conjunctive semantics
    await DB.insert(content='Text', title='Some text', page_numbers=[1], tags=['tag1'])
    await DB.insert(content='Text', title='Some mo text', page_numbers=[1], tags=['tag2', 'tag3'])
    await DB.insert(content='Text', title='Even mo text', page_numbers=[1], tags=['tag3'])

    # Using limit default
    sim_search = await DB.search(text='Text', tags=['tag1', 'tag3'], conjunctive=False)
    assert sim_search is not None, Exception("No results returned from filtered search")
    assert len(list(sim_search)) == 3, Exception(f"There should be 3 results, received {sim_search}")

    sim_search = await DB.search(text='Text', tags=['tag1', 'tag3'], conjunctive=False, limit=1000)
    assert sim_search is not None, Exception("No results returned from filtered search")
    assert len(list(sim_search)) == 3, Exception(f"There should be 3 results, received {sim_search}")

    texts = ['Hello world', 'Hello Dolly', 'Good-Bye to All That']
    authors = ['Brian Kernighan', 'Louis Armstrong', 'Robert Graves']
    metas = [[f'author={a}'] for a in authors]
    count = len(texts)
    records = zip(texts, metas, ['']*count, [None]*count)
    await DB.insert_many(records)

    sim_search = await DB.search(text='Hi there!', threshold=0.999, limit=0)
    assert sim_search is not None, Exception("No results returned from filtered search")
    assert len(list(sim_search)) == 3, Exception(f"There should be 3 results, received {sim_search}")

    sim_search = await DB.search(text='Hi there!', threshold=0.999, limit=2)
    assert sim_search is not None, Exception("No results returned from filtered search")
    assert len(list(sim_search)) == 2, Exception(f"There should be 2 results, received {sim_search}")

    await DB.drop_table()


if __name__ == '__main__':
    raise SystemExit("Attention! Run with pytest")



'''
pytest test

or

pytest test/test_orca_prompting.py
'''
# import pytest

from ogbujipt.prompting.basic import format
from ogbujipt.prompting.model_style import ORCA_DELIMITERS


def orca_prompt_building(PROMPTING_USER_QUERY):
    EXPECTED_PROMPT = '### System:\nYou are an AI assistant that follows instruction extremely well. Help as much as you can.\n\n### User:\nwhat\'s nine plus ten?\n\n### Response:\n'  # noqa: E501

    prompt = format(
        PROMPTING_USER_QUERY,
        delimiters=ORCA_DELIMITERS
        )
    assert prompt == EXPECTED_PROMPT


def orca_input_prompt_building(PROMPTING_USER_QUERY, PROMPTING_CONTEXT):
    EXPECTED_PROMPT = '### System:\nYou are an AI assistant that follows instruction extremely well. Help as much as you can.\n\n### User:\nwhat\'s nine plus ten?\n\n### Input:\nThe entirety of "Principia Mathematica" by Isaac Newton\n\n### Response:\n'  # noqa: E501

    prompt = format(
        PROMPTING_USER_QUERY,
        contexts=PROMPTING_CONTEXT,
        delimiters=ORCA_DELIMITERS
        )
    assert prompt == EXPECTED_PROMPT



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# test/embedding/test_pgvector_message.py
'''
See test/embedding/test_pgvector.py for important notes on running these tests
'''

from datetime import datetime

import pytest
from unittest.mock import MagicMock, DEFAULT  # noqa: F401

# import numpy as np

# XXX: Move to a fixture?
# Definitely don't want to even import SentenceTransformer class due to massive side-effects
class SentenceTransformer(object):
    def __init__(self, model_name_or_path):
        self.encode = MagicMock()


@pytest.fixture
def MESSAGES():
    messages = [  # Test data: history_key, role, content, timestamp, metadata
    ('00000000-0000-0000-0000-000000000000', 'ama', 'Hello Eme!', '2021-10-01 00:00:00+00:00', {'1': 'a'}),
    ('00000000-0000-0000-0000-000000000001', 'ugo', 'Greetings Ego', '2021-10-01 00:00:01+00:00', {'2': 'b'}),
    ('00000000-0000-0000-0000-000000000000', 'eme', 'How you dey, Ama!', '2021-10-01 00:00:02+00:00', {'3': 'c'}),
    ('00000000-0000-0000-0000-000000000001', 'ego', 'What a pleasant surprise', '2021-10-01 00:00:03+00:00', {'4': 'd'}),  # noqa: E501
    ('00000000-0000-0000-0000-000000000000', 'ama', 'Not bad, not bad at all', '2021-10-01 00:00:04+00:00', {'5': 'e'}),  # noqa: E501
    ('00000000-0000-0000-0000-000000000001', 'ugo', 'Glad you think so. I was planning to drop by later', '2021-10-01 00:00:05+00:00', {'6': 'f'}),  # noqa: E501
    ('00000000-0000-0000-0000-000000000000', 'eme', 'Very good. Say hello to your family for me.', '2021-10-01 00:00:06+00:00', {'7': 'g'}),  # noqa: E501
    ('00000000-0000-0000-0000-000000000001', 'ugo', 'An even better surprise, I hope!', '2021-10-01 00:00:07+00:00', {'8': 'h'})  # noqa: E501
    ]
    return [(history_key, role, content, datetime.fromisoformat(timestamp), metadata)
    for history_key, role, content, timestamp, metadata in messages
    ]


@pytest.mark.asyncio
async def test_insert_message_vector(DB, MESSAGES):
    # Insert data
    for index, (row) in enumerate(MESSAGES):
        await DB.insert(*row)

    assert await DB.count_items() == len(MESSAGES), Exception('Incorrect number of messages after insertion')

    history_key, role, content, timestamp, metadata = MESSAGES[0]

    # search table with perfect match
    result = await DB.search(text=content, history_key=history_key, limit=3)
    # assert result is not None, Exception('No results returned from perfect search')

    # Even though the embedding is mocked, the stored text should be faithful
    row = next(result)
    assert row.content == content
    assert row.metadata == {'1': 'a'}


@pytest.mark.asyncio
async def test_insertmany_message_vector(DB, MESSAGES):
    # Insert data using insert_many()
    await DB.insert_many(MESSAGES)

    assert await DB.count_items() == len(MESSAGES), Exception('Incorrect number of messages after insertion')

    history_key, role, content, timestamp, metadata = MESSAGES[0]

    # search table with perfect match
    result = await DB.search(text=content, history_key=history_key, limit=3)
    # assert result is not None, Exception('No results returned from perfect search')

    # Even though the embedding is mocked, the stored text should be faithful
    row = next(result)
    assert row.content == content
    assert row.metadata == {'1': 'a'}


@pytest.mark.asyncio
async def test_get_messages_all_limit(DB, MESSAGES):
    # Insert data using insert_many()
    await DB.insert_many(MESSAGES)

    history_key, role, content, timestamp, metadata = MESSAGES[0]

    results = await DB.get_messages(history_key=history_key)
    assert len(list(results)) == 4, Exception('Incorrect number of messages returned from chatlog')

    results = await DB.get_messages(history_key=history_key, limit=3)
    assert len(list(results)) == 3, Exception('Incorrect number of messages returned from chatlog')

    # With limit, should return the most recent messages
    history_key, role, content, timestamp, metadata = MESSAGES[-1]

    results = list(await DB.get_messages(history_key=history_key, limit=1))
    assert len(results) == 1, Exception('Incorrect number of messages returned from chatlog')
    assert results[0].content == content, Exception('Incorrect message returned from chatlog')


@pytest.mark.asyncio
async def test_get_messages_since(DB, MESSAGES):
    await DB.insert_many(MESSAGES)

    history_key, role, content, timestamp, metadata = MESSAGES[0]

    since_ts = datetime.fromisoformat('2021-10-01 00:00:03+00:00')
    results = list(await DB.get_messages(history_key=history_key, since=since_ts))
    assert len(results) == 2, Exception('Incorrect number of messages returned from chatlog')

    since_ts = datetime.fromisoformat('2021-10-01 00:00:04+00:00')
    results = list(await DB.get_messages(history_key=history_key, since=since_ts))
    assert len(results) == 1, Exception('Incorrect number of messages returned from chatlog')


if __name__ == '__main__':
    raise SystemExit("Attention! Run with pytest")



'''
Downloads models from Hugging Face to models/model-name.

Examples:
python download-model.py facebook/opt-1.3b

Using the built-in short name:
python download-model.py "OPT 1.3B"

Specifying a subset of model files within a repository folder
python download-model.py --output . --select="llongorca-13b-16k.Q4_K_M.gguf|llongorca-13b-16k.Q5_K_M.gguf" TheBloke/LlongOrca-13B-16K-GGUF

Fork of https://github.com/oobabooga/text-generation-webui/blob/main/download-model.py

Adds --select option to download-model.py
Also some code clean-up, e.g. avoiding shadowing builtins (i.e. dict & bytes)
See also: https://github.com/oobabooga/text-generation-webui/pull/2508
'''  # noqa: E501

import argparse
import base64
import datetime
import hashlib
import re
import os
import sys
from pathlib import Path

import requests
import tqdm
from tqdm.contrib.concurrent import thread_map


def select_model_from_default_options():
    models = {
        "OPT 6.7B": ("facebook", "opt-6.7b", "main"),
        "OPT 2.7B": ("facebook", "opt-2.7b", "main"),
        "OPT 1.3B": ("facebook", "opt-1.3b", "main"),
        "OPT 350M": ("facebook", "opt-350m", "main"),
        "GALACTICA 6.7B": ("facebook", "galactica-6.7b", "main"),
        "GALACTICA 1.3B": ("facebook", "galactica-1.3b", "main"),
        "GALACTICA 125M": ("facebook", "galactica-125m", "main"),
        "Pythia-6.9B-deduped": ("EleutherAI", "pythia-6.9b-deduped", "main"),
        "Pythia-2.8B-deduped": ("EleutherAI", "pythia-2.8b-deduped", "main"),
        "Pythia-1.4B-deduped": ("EleutherAI", "pythia-1.4b-deduped", "main"),
        "Pythia-410M-deduped": ("EleutherAI", "pythia-410m-deduped", "main"),
    }

    choices = {}
    print("Select the model that you want to download:\n")
    for i, name in enumerate(models):
        char = chr(ord('A') + i)
        choices[char] = name
        print(f"{char}) {name}")

    char_hugging = chr(ord('A') + len(models))
    print(f"{char_hugging}) Manually specify a Hugging Face model")
    char_exit = chr(ord('A') + len(models) + 1)
    print(f"{char_exit}) Do not download a model")
    print()
    print("Input> ", end='')
    choice = input()[0].strip().upper()
    if choice == char_exit:
        exit()
    elif choice == char_hugging:
        print("""\nType the name of your desired Hugging Face model in the format organization/name.

Examples:
facebook/opt-1.3b
EleutherAI/pythia-1.4b-deduped
""")

        print("Input> ", end='')
        model = input()
        branch = "main"
    else:
        arr = models[choices[choice]]
        model = f"{arr[0]}/{arr[1]}"
        branch = arr[2]

    return model, branch


class ModelDownloader:
    """
    Manages state for downloading from HF, especially with regard to auth,
    allowing download of protected models
    """
    def __init__(self):
        self._session = requests.Session()
        if all((os.getenv('HF_USER'), os.getenv('HF_PASS'))):
            self._session.auth = (os.getenv('HF_USER'), os.getenv('HF_PASS'))

    def sanitize_model_and_branch_names(self, model, branch):
        """
        Make sure model & branch names are valid & do some clean-up, if need be
        """
        model = model.rstrip("/")

        if branch is None:
            branch = "main"
        else:
            pattern = re.compile(r"^[a-zA-Z0-9._-]+$")
            if not pattern.match(branch):
                raise ValueError(
                    "Invalid branch name. Only alphanumeric characters, period, underscore and dash are allowed.")

        return model, branch

    def get_download_links_from_huggingface(self, model, branch, text_only=False, select=None):
        """
        Gather a list of download links & checksums (SHA256) from all pages of the files manifest for the HF repository
        """
        base = "https://huggingface.co"
        page = f"/api/models/{model}/tree/{branch}"
        cursor = b""

        if select:
            select = select.split("|")

        links = []
        sha256 = []
        classifications = []
        has_pytorch = False
        has_pt = False
        has_ggml = False
        has_gguf = False
        has_safetensors = False
        is_lora = False
        # The loop is for paging over repos with many files
        while True:
            # Get the file manifest from the HF repo
            url = f"{base}{page}" + (f"?cursor={cursor.decode()}" if cursor else "")
            resp = self._session.get(url, timeout=10)
            resp.raise_for_status()
            manifest = resp.json()

            # We've hit an empty page, so we're done
            if len(manifest) == 0:
                break

            for i in range(len(manifest)):
                fname = manifest[i]['path']
                if not is_lora and fname.endswith(('adapter_config.json', 'adapter_model.bin')):
                    is_lora = True

                is_pytorch = re.match("(pytorch|adapter)_model.*\.bin", fname)
                is_safetensors = re.match(".*\.safetensors", fname)
                is_pt = re.match(".*\.pt", fname)
                is_ggml = re.match(".*ggml.*\.bin", fname)
                is_gguf = re.match(".*gguf", fname)
                is_tokenizer = re.match("(tokenizer|ice).*\.model", fname)
                is_text = re.match(".*\.(txt|json|py|md)", fname) or is_tokenizer
                if any((is_pytorch, is_safetensors, is_pt, is_ggml, is_gguf, is_tokenizer, is_text)):
                    # if a select option is given, and current file is a binary, make sure it's in the select list
                    if select and any((is_pytorch, is_pt, is_safetensors, is_ggml, is_gguf)):
                        if fname not in select:
                            continue

                    if 'lfs' in manifest[i]:
                        sha256.append([fname, manifest[i]['lfs']['oid']])

                    if is_text:
                        links.append(f"https://huggingface.co/{model}/resolve/{branch}/{fname}")
                        classifications.append('text')
                        continue

                    if not text_only:
                        links.append(f"https://huggingface.co/{model}/resolve/{branch}/{fname}")
                        if is_safetensors:
                            has_safetensors = True
                            classifications.append('safetensors')
                        elif is_pytorch:
                            has_pytorch = True
                            classifications.append('pytorch')
                        elif is_pt:
                            has_pt = True
                            classifications.append('pt')
                        elif is_ggml:
                            has_ggml = True  # noqa F841 (this will eventually be a useful flag)
                            classifications.append('ggml')
                        elif is_gguf:
                            has_gguf = True  # noqa F841 (this will eventually be a useful flag)
                            classifications.append('gguf')

            # Set up to move the next page (if any) of HF repo contents
            cursor = base64.b64encode(f'{{"file_name":"{manifest[-1]["path"]}"}}'.encode()) + b':50'
            cursor = base64.b64encode(cursor)
            cursor = cursor.replace(b'=', b'%3D')

        # If both pytorch and safetensors are available, download safetensors only
        if (has_pytorch or has_pt) and has_safetensors:
            for i in range(len(classifications) - 1, -1, -1):
                if classifications[i] in ['pytorch', 'pt']:
                    links.pop(i)

        return links, sha256, is_lora

    def get_output_folder(self, model, branch, is_lora, base_folder=None):
        """
        Figure out where to put the downloaded files, based on user params,
        model type & HF repo branch
        """
        if base_folder is None:
            base_folder = 'models' if not is_lora else 'loras'

        output_folder = f"{'_'.join(model.split('/')[-2:])}"
        if branch != 'main':
            output_folder += f'_{branch}'
        output_folder = Path(base_folder) / output_folder
        return output_folder

    def get_single_file(self, url, output_folder, start_from_scratch=False):
        """
        Download an individual file from the HF repo, within a thread. Also handles resume.
        """
        filename = Path(url.rsplit('/', 1)[1])
        output_path = output_folder / filename
        if output_path.exists() and not start_from_scratch:
            # Check if the file has already been downloaded completely
            resp = self._session.get(url, stream=True, timeout=10)
            total_size = int(resp.headers.get('content-length', 0))
            if output_path.stat().st_size >= total_size:
                return
            # Otherwise, resume the download from where it left off
            headers = {'Range': f'bytes={output_path.stat().st_size}-'}
            mode = 'ab'
        else:
            headers = {}
            mode = 'wb'

        resp = self._session.get(url, stream=True, headers=headers, timeout=10)
        with open(output_path, mode) as f:
            total_size = int(resp.headers.get('content-length', 0))
            block_size = 1024
            with tqdm.tqdm(total=total_size, unit='iB', unit_scale=True, 
                           bar_format='{l_bar}{bar}| {n_fmt:6}/{total_fmt:6} {rate_fmt:6}') as t:
                for data in resp.iter_content(block_size):
                    t.update(len(data))
                    f.write(data)

    def start_download_threads(self, file_list, output_folder, start_from_scratch=False, threads=1):
        thread_map(lambda url: self.get_single_file(url, output_folder, start_from_scratch=start_from_scratch), 
                   file_list, max_workers=threads, disable=True)

    def download_model_files(self, model, branch, links, sha256, output_folder, start_from_scratch=False, threads=1):
        """
        Make the preparations to launch the threads to download the individual files
        """
        # Creating the folder and writing the metadata
        if not output_folder.exists():
            output_folder.mkdir(parents=True, exist_ok=True)
        with open(output_folder / 'huggingface-metadata.txt', 'w') as f:
            f.write(f'url: https://huggingface.co/{model}\n')
            f.write(f'branch: {branch}\n')
            f.write(f'download date: {str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"))}\n')
            sha256_str = ''
            for i in range(len(sha256)):
                sha256_str += f'    {sha256[i][1]} {sha256[i][0]}\n'
            if sha256_str != '':
                f.write(f'sha256sum:\n{sha256_str}')

        # Downloading the files
        print(f"Downloading the model to {output_folder}")
        self.start_download_threads(links, output_folder, start_from_scratch=start_from_scratch, threads=threads)

    def check_model_files(self, model, branch, links, sha256, output_folder):
        """
        Validate the checksums of model files, to make sure we completely downloaded exactly what we should have
        """
        # Validate the checksums
        validated = True
        for i in range(len(sha256)):
            fpath = (output_folder / sha256[i][0])

            if not fpath.exists():
                print(f"The following file is missing: {fpath}")
                validated = False
                continue

            with open(output_folder / sha256[i][0], "rb") as f:
                fbytes = f.read()
                file_hash = hashlib.sha256(fbytes).hexdigest()
                if file_hash != sha256[i][1]:
                    print(f'Checksum failed: {sha256[i][0]}  {sha256[i][1]}')
                    validated = False
                else:
                    print(f'Checksum validated: {sha256[i][0]}  {sha256[i][1]}')

        if validated:
            print('[+] Validated checksums of all model files!')
        else:
            print('[-] Invalid checksums. Rerun download-model.py with the --clean flag.')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('MODEL', type=str, default=None, nargs='?')
    parser.add_argument('--branch', type=str, default='main', help='Name of the Git branch to download from.')
    parser.add_argument('--threads', type=int, default=1, help='Number of files to download simultaneously.')
    parser.add_argument('--text-only', action='store_true', help='Only download text files (txt/json).')
    parser.add_argument('--output', type=str, default=None, help='The folder where the model should be saved.')
    parser.add_argument('--clean', action='store_true', help='Does not resume the previous download.')
    parser.add_argument('--check', action='store_true', help='Validates the checksums of model files.')
    parser.add_argument('--select', type=str, help='Pipe separated list of .bin or .safetensors files to download. '\
                        'No other files of these extensions will be.')
    args = parser.parse_args()

    branch = args.branch
    model = args.MODEL
    if model is None:
        model, branch = select_model_from_default_options()

    downloader = ModelDownloader()
    # Cleaning up the model/branch names
    try:
        model, branch = downloader.sanitize_model_and_branch_names(model, branch)
    except ValueError as err_branch:
        print(f"Error: {err_branch}")
        sys.exit()

    # Getting the download links from Hugging Face
    links, sha256, is_lora = downloader.get_download_links_from_huggingface(model, branch, text_only=args.text_only, 
                                                                            select=args.select)

    # Getting the output folder
    output_folder = downloader.get_output_folder(model, branch, is_lora, base_folder=args.output)

    if args.check:
        # Check previously downloaded files
        downloader.check_model_files(model, branch, links, sha256, output_folder)
    else:
        # Download files
        downloader.download_model_files(model, branch, links, sha256, output_folder, threads=args.threads)



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# ogbujipt/demo/multiprocess.py
'''
Advanced demo showing quick chat with LLMs, with 3 simultaneous requests.
It can be configured to chat with 3 separate LLMs, or 3 instances of the same LLM.
Includes a progress indicator dislay while the LLM instances are generating.
Takes advantage of Python's asyncio, and also multiprocess, which requires some finesse,
mostly handled by OgbujiPT. Works even when the LLM host in use doesn't suport asyncio.

```sh
python demo/multiprocess.py --apibase=http://my-llm-host:8000
```

Also allows you to use the actual OpenAI ChatGPT service, by specifying --openai
'''
import asyncio
from pathlib import Path

import click

from ogbujipt import word_loom
from ogbujipt.llm_wrapper import openai_chat_api, prompt_to_chat
from ogbujipt.async_helper import console_progress_indicator


def file_path_here():
    '''Cross-platform Python trick to get the path to this very file'''
    from inspect import getsourcefile
    from os.path import abspath

    return abspath(getsourcefile(lambda: 0))


# Load language material, such as prompt templates, from an external file—word loom format
HERE = Path(file_path_here()).parent
with open(HERE / Path('language.toml'), mode='rb') as fp:
    TEXTS = word_loom.load(fp)


async def async_main(requests_info):
    '''
    Main entry point for asyncio
    model_topics - list of (llm_wrapped, topic, temperature) tuples
    llmtemp - LLM temperature; hard-coded across all (see https://beta.openai.com/docs/api-reference/completions/create)
    '''
    # Pro tip: When creating asyncio tasks be mindful to not lose references to tasks,
    # lest they get garbage collected, which sows chaos. asyncio.TaskGroup (new in Python 3.11)
    # is often a better alternative, but waits for all tasks to complete whereas we're done once
    # the LLM generation tasks are complete
    indicator_task = asyncio.create_task(console_progress_indicator())
    llm_tasks = [llm.wrap_for_multiproc(prompt_to_chat(msg), temperature=temp, max_tokens=1024)
                 for (llm, msg, temp) in requests_info]
    llm_messages = [msg for (llm, msg, temp) in requests_info]
    # Need to gather to make sure all LLM tasks are completed
    gathered_llm_tasks = asyncio.gather(*llm_tasks)
    done, _ = await asyncio.wait((indicator_task, gathered_llm_tasks), return_when=asyncio.FIRST_COMPLETED)

    # Completed task will from gather() of llm_tasks; results in original task arg order
    results = zip(llm_messages, next(iter(done)).result())
    for msg, resp in results:
        print(f'Result re {msg}')
        # resp is an instance of openai.openai_object.OpenAIObject, with lots of useful info
        print('\nFull response data from LLM:\n', resp)
        # Just the response text
        response_text = openai_chat_api.first_choice_message(resp)
        print('\nResponse text from LLM:\n\n', response_text)
        print('-'*80)


# Command line arguments defined in click decorators
@click.command()
@click.option('--apibase', default='http://127.0.0.1:8000', help='OpenAI API base URL')
@click.option('--llmtemp', default='0.9', type=float, help='LLM temperature')
@click.option('--openai', is_flag=True, default=False, type=bool,
              help='Use live OpenAI API. If you use this option, you must have '
              '"OPENAI_API_KEY" defined in your environmnt')
@click.option('--model', default='', type=str, 
              help='OpenAI model to use (see https://platform.openai.com/docs/models)')
def main(apibase, llmtemp, openai, model):
    # Use OpenAI API if specified, otherwise emulate with supplied URL info
    if openai:
        oapi = openai_chat_api(model=(model or 'gpt-3.5-turbo'))
    else:
        oapi = openai_chat_api(model=model, api_base=apibase)

    # Separate models or params—e.g. temp—for each LLM request is left as an exercise 😊
    requests_info = [
        (oapi, TEXTS['test_prompt_joke'].format(topic=t), llmtemp)
        for t in ('wild animals', 'vehicles', 'space aliens')]

    asyncio.run(async_main(requests_info))


if __name__ == '__main__':
    # CLI entry point. Also protects against re-execution of main() after process fork
    # viz https://docs.python.org/3/library/multiprocessing.html#multiprocessing-safe-main-import
    main()



'''
Common tools & resources for test cases
'''
# ruff: noqa: E501

import pytest
from openai.types.completion import Completion, CompletionChoice, CompletionUsage
from openai.types.chat.chat_completion import ChatCompletion, Choice, ChatCompletionMessage


@pytest.fixture
def PROMPTING_USER_QUERY():
    return 'what\'s nine plus ten?'


@pytest.fixture
def PROMPTING_CONTEXT():
    return 'The entirety of "Principia Mathematica" by Isaac Newton'


@pytest.fixture
def PROMPTING_CONTEXT_LIST():
    return [
        (
            'author: Isaac Newton\n' 'title: Principia Mathematica Volume 1\n' 'date: 1910\n' 'ISBN-13: 978-0521626063',
            'All mathematical propositions\' are of the form S is P\'; S\' is the subject, P\' the predicate.',
        ),
        (
            'author: Isaac Newton\n' 'title: Principia Mathematica Volume 2\n' 'date: 1912\n' 'ISBN-13: 978-0521626070',
            'It is desirable to find some convenient phrase which shall denote the whole collection of values of a function, or the whole collection of entities of a given type.',
        ),
        (
            'author: Isaac Newton\n' 'title: Principia Mathematica Volume 3\n' 'date: 1913\n' 'ISBN-13: 978-0521626087',
            'In this volume we shall define the arithmetical relations between finite cardinal numbers and also their addition and multiplication.',
        ),
        (
            'author: Isaac Newton\n' 'title: Principia Mathematica Volume 4\n' 'date: 1915\n' 'ISBN-13: 978-6969696969',
            'Finally, we can get onto the real purpose of this work; to explain why 9 + 10 = 21',
        ),
    ]


@pytest.fixture
def BAD_XML_CODE():
    return '''\
<earth>
<country><b>Russia</country></b>
<capital>Moscow</capital>
</Earth>'''


@pytest.fixture
def COME_THUNDER_POEM():
    # One of Christopher Okigbo's greatest poems
    return '''\
Now that the triumphant march has entered the last street corners,
Remember, O dancers, the thunder among the clouds…

Now that the laughter, broken in two, hangs tremulous between the teeth,
Remember, O Dancers, the lightning beyond the earth…

The smell of blood already floats in the lavender-mist of the afternoon.
The death sentence lies in ambush along the corridors of power;
And a great fearful thing already tugs at the cables of the open air,
A nebula immense and immeasurable, a night of deep waters —
An iron dream unnamed and unprintable, a path of stone.

The drowsy heads of the pods in barren farmlands witness it,
The homesteads abandoned in this century’s brush fire witness it:
The myriad eyes of deserted corn cobs in burning barns witness it:
Magic birds with the miracle of lightning flash on their feathers…

The arrows of God tremble at the gates of light,
The drums of curfew pander to a dance of death;

And the secret thing in its heaving
Threatens with iron mask
The last lighted torch of the century…'''


@pytest.fixture
def COME_THUNDER_POEM_CHUNKS():
    return [
        'Now that the triumphant march has entered the last street corners,\nRemember, O dancers, the thunder among the clouds…\n\nNow that the laughter, broken in two, hangs tremulous between the teeth,\nRemember, O Dancers, the lightning beyond the earth…',
        'Now that the triumphant march has entered the last street corners,\nRemember, O dancers, the thunder among the clouds…\n\nNow that the laughter, broken in two, hangs tremulous between the teeth,\nRemember, O Dancers, the lightning beyond the earth…\n\nThe smell of blood already floats in the lavender-mist of the afternoon.\nThe death sentence lies in ambush along the corridors of power;\nAnd a great fearful thing already tugs at the cables of the open air,\nA nebula immense and immeasurable, a night of deep waters —\nAn iron dream unnamed and unprintable, a path of stone.',
        'Now that the laughter, broken in two, hangs tremulous between the teeth,\nRemember, O Dancers, the lightning beyond the earth…\n\nThe smell of blood already floats in the lavender-mist of the afternoon.\nThe death sentence lies in ambush along the corridors of power;\nAnd a great fearful thing already tugs at the cables of the open air,\nA nebula immense and immeasurable, a night of deep waters —\nAn iron dream unnamed and unprintable, a path of stone.\n\nThe drowsy heads of the pods in barren farmlands witness it,\nThe homesteads abandoned in this century’s brush fire witness it:\nThe myriad eyes of deserted corn cobs in burning barns witness it:\nMagic birds with the miracle of lightning flash on their feathers…',
        'The smell of blood already floats in the lavender-mist of the afternoon.\nThe death sentence lies in ambush along the corridors of power;\nAnd a great fearful thing already tugs at the cables of the open air,\nA nebula immense and immeasurable, a night of deep waters —\nAn iron dream unnamed and unprintable, a path of stone.\n\nThe drowsy heads of the pods in barren farmlands witness it,\nThe homesteads abandoned in this century’s brush fire witness it:\nThe myriad eyes of deserted corn cobs in burning barns witness it:\nMagic birds with the miracle of lightning flash on their feathers…\n\nThe arrows of God tremble at the gates of light,\nThe drums of curfew pander to a dance of death;',
    ]


@pytest.fixture
def BASIC_EVEN_BLOCK():
    # Simple example of text with even-length lines
    return '''\
0123456789
abcdefghij
ABCDEFGHIJ
klmnopqrst'''


@pytest.fixture
def LOREM_IPSUM():
    # Lorem ipsum text
    return (
        'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Fusce vestibulum nisl eget mauris malesuada, '
        'quis facilisis arcu vehicula. Sed consequat, quam ut auctor volutpat, augue ex tincidunt massa, in varius '
        'nulla ex vel ipsum. Nullam vitae eros nec ante sagittis luctus. Nullam scelerisque dolor eu orci iaculis, '
        'at convallis nulla luctus. Praesent eget ex id arcu facilisis varius vel id neque. Donec non orci eget '
        'elit aliquam tempus. Sed at tortor at tortor congue dictum. Nulla varius erat at libero lacinia, id '
        'dignissim risus auctor. Ut eu odio vehicula, tincidunt justo ac, viverra erat. Sed nec sem sit amet erat '
        'malesuada finibus. Nulla sit amet diam nec dolor tristique dignissim. Sed vehicula, justo nec posuere '
        'eleifend, libero ligula interdum neque, at lacinia arcu quam non est. Integer aliquet, erat id dictum '
        'euismod, felis libero blandit lorem, nec ullamcorper quam justo at elit.'
    )


@pytest.fixture
def OPENAI_TEXT_RESPONSE_OBJECT():
    # Example of OpenAI response object
    # Response generated by prompting with "Recite the Fitness Gram Pacer Test script."
    return Completion(
        id='cmpl-13ca5e85-a890-4c70-bbbf-9f4825728f23',
        choices=[
            CompletionChoice(
                finish_reason='length',
                index=0,
                logprobs=None,
                text='…is an exceptional employee who has made significant contributions to our company.',
            )
        ],
        created=1701544442,
        model='.local/share/models/TheBloke_Llama-2-7B-32K-Instruct-GGUF/llama-2-7b-32k-instruct.Q5_K_M.gguf',
        object='text_completion',
        system_fingerprint=None,
        usage=CompletionUsage(completion_tokens=256, prompt_tokens=12, total_tokens=268),
    )


# TODO: Finish up, then enable test in test/test_ogbujipt.py
@pytest.fixture
def OPENAI_MSG_RESPONSE_OBJECT():
    return ChatCompletion(
        id='chatcmpl-923076c2-d1d5-473b-bdc6-6e803037357a',
        choices=[
            Choice(
                finish_reason='length',
                index=0,
                message=ChatCompletionMessage(
                    content='…is an exceptional employee who has made significant contributions to our company.',
                    role='assistant',
                    function_call=None,
                    tool_calls=None,
                ),
                logprobs=None
            )
        ],
        created=1701544162,
        model='.local/share/models/TheBloke_Llama-2-7B-32K-Instruct-GGUF/llama-2-7b-32k-instruct.Q5_K_M.gguf',
        object='chat.completion',
        system_fingerprint=None,
        usage=CompletionUsage(completion_tokens=256, prompt_tokens=22, total_tokens=278),
    )



'''
pytest test

or

pytest test/test_alpaca_prompting.py
'''
# import pytest

from ogbujipt.prompting.basic import format
from ogbujipt.prompting.model_style import (
    ALPACA_DELIMITERS,
    ALPACA_INSTRUCT_INPUT_DELIMITERS
    )

def alpaca_prompt_building(PROMPTING_USER_QUERY):
    EXPECTED_PROMPT = 'what\'s nine plus ten?\n\n### Response:\n'  # noqa: E501

    prompt = format(
        PROMPTING_USER_QUERY,
        delimiters=ALPACA_DELIMITERS
        )
    assert prompt == EXPECTED_PROMPT


def alpaca_instruct_prompt_building(PROMPTING_USER_QUERY):
    EXPECTED_PROMPT = '### Instruction:\nwhat\'s nine plus ten?\n\n### Response:\n'  # noqa: E501

    prompt = format(
        PROMPTING_USER_QUERY,
        delimiters=ALPACA_INSTRUCT_INPUT_DELIMITERS
        )
    assert prompt == EXPECTED_PROMPT


def alpaca_instruct_input_prompt_building(PROMPTING_USER_QUERY, PROMPTING_CONTEXT):
    EXPECTED_PROMPT = '### Instruction:\nwhat\'s nine plus ten?\n\n### Input:\nThe entirety of "Principia Mathematica" by Isaac Newton\n\n### Response:\n'  # noqa: E501

    prompt = format(
        PROMPTING_USER_QUERY,
        contexts=PROMPTING_CONTEXT,
        delimiters=ALPACA_INSTRUCT_INPUT_DELIMITERS
        )
    assert prompt == EXPECTED_PROMPT

